/* oe_basic.c -- Basic matrix functions for optimal estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "oe.h"

/* Set elements of a vector to zero */
int
oe_zero_vector(OEvector *v)
{
  if (!v->data) {
    OE_THROW(OE_DIMENSION_ERROR, "Cannot zero unallocated vector");
    return OE_FAILURE;
  }
  else {
    int j;
    for (j = 0; j < v->n; j++) {
      v->data[j] = 0.0;
    }
    return OE_SUCCESS;
  }
}

/* Set elements of an integer vector to zero */
int
oe_zero_int_vector(OEintvector *v)
{
  if (!v->data) {
    OE_THROW(OE_DIMENSION_ERROR, "Cannot zero unallocated vector");
    return OE_FAILURE;
  }
  else {
    int j;
    for (j = 0; j < v->n; j++) {
      v->data[j] = 0;
    }
    return OE_SUCCESS;
  }
}

/* Set elements of a matrix to zero */
int
oe_zero_matrix(OEmatrix *M)
{
  if (!M->data) {
    OE_THROW(OE_DIMENSION_ERROR, "Cannot zero unallocated matrix");
  }
  else {
    int i, j;
    for (j = 0; j < M->m; j++) {
      for (i = 0; i < M->n; i++) {
	M->data[j][i] = 0.0;
      }
    }
    return OE_SUCCESS;
  }
}

/* Set elements of an integer matrix to zero */
int
oe_zero_int_matrix(OEintmatrix *M)
{
  if (!M->data) {
    OE_THROW(OE_DIMENSION_ERROR, "Cannot zero unallocated matrix");
  }
  else {
    int i, j;
    for (j = 0; j < M->m; j++) {
      for (i = 0; i < M->n; i++) {
	M->data[j][i] = 0;
      }
    }
    return OE_SUCCESS;
  }
}

/* Set elements of a vector to a constant */
int
oe_constant_vector(OEvector *v, OEreal value)
{
  if (!v->data) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Cannot set elements of an unallocated vector");
  }
  else {
    int j;
    for (j = 0; j < v->n; j++) {
      v->data[j] = value;
    }
    return OE_SUCCESS;
  }
}

/* Set elements of a matrix to a constant */
int
oe_constant_matrix(OEmatrix *M, OEreal value)
{
  if (!M->data) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Cannot set elements of an unallocated matrix");
  }
  else {
    int i, j;
    for (j = 0; j < M->m; j++) {
      for (i = 0; i < M->n; i++) {
	M->data[j][i] = value;
      }
    }
    return OE_SUCCESS;
  }
}

/* Create a copy of vector v_src in v_dest */
int
oe_copy_vector(OEvector *v_src, OEvector *v_dest)
{
  if (v_src->n != v_dest->n) {
    OE_CHECK(oe_reset_vector(v_dest, v_src->n));
  }
  memcpy(v_dest->data, v_src->data, v_src->n * sizeof(OEreal));
  return OE_SUCCESS;
}

/* Copy part of v_src to memory pointed to by dest */
int
oe_copy_vector_part(OEvector *v_src, int src_start, int num,
		    OEreal *dest)
{
  if (src_start < 0 || num < 0 || src_start+num > v_src->n) {
    OE_THROW(OE_DIMENSION_ERROR, "Vector range lies outside vector");
  }
  memcpy(dest, v_src->data+src_start, num * sizeof(OEreal));
  return OE_SUCCESS;
}

/* Create a copy of matrix M_src in M_dest */
int
oe_copy_matrix(OEmatrix *M_src, OEmatrix *M_dest)
{
  int j;
  if (M_src->n != M_dest->n || M_src->m != M_dest->m) {
    OE_CHECK(oe_reset_matrix(M_dest, M_src->m, M_src->n));
  }
  for (j = 0; j < M_src->m; j++) {
    memcpy(M_dest->data[j], M_src->data[j], 
	   M_src->n * sizeof(OEreal));
  }
  return OE_SUCCESS;
}

/* Copy part of matrix M_src to part of M_dest */
int
oe_copy_matrix_part(OEmatrix *M_src,
		    int src_m_start, int src_n_start,
		    int num_m, int num_n,
		    OEmatrix *M_dest,
		    int dest_m_start, int dest_n_start)
{
  int j;
  if (num_m < 0 || num_n < 0
      || src_m_start < 0 || src_n_start < 0
      || dest_m_start < 0 || dest_n_start < 0
      || src_m_start+num_m > M_src->m
      || src_n_start+num_n > M_src->n
      || dest_m_start+num_m > M_dest->m
      || dest_n_start+num_n > M_dest->n) {
    OE_THROW(OE_DIMENSION_ERROR, "Part of matrix to be copied lies outside source or destination matrix");
  }

  for (j = src_m_start; j < src_m_start+num_m; j++) {
    memcpy(M_dest->data[j]+dest_n_start, M_src->data[j]+src_n_start, 
	   num_n * sizeof(OEreal));
  }
  return OE_SUCCESS;
}

/* Reverse the elements of a vector */
int
oe_reverse_vector(OEvector *v)
{
  OEreal val;
  int i, l = v->n / 2;
  for (i = 0; i < l; i++) {
    val = v->data[i];
    v->data[i] = v->data[v->n - i - 1];
    v->data[v->n - i - 1] = val;
  }
  return OE_SUCCESS;
}

/* Scale elements of a vector by a constant factor */
int
oe_scale_vector(OEvector *v, OEreal scaling)
{
  if (!v->data) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Cannot scale elements of an unallocated vector");
  }
  else {
    int j;
    for (j = 0; j < v->n; j++) {
      v->data[j] *= scaling;
    }
    return OE_SUCCESS;
  }
}

/* Scale elements of a matrix by a constant factor */
int
oe_scale_matrix(OEmatrix *M, OEreal scaling)
{
  if (!M->data) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Cannot scale elements of an unallocated matrix");
  }
  else {
    int i, j;
    for (j = 0; j < M->m; j++) {
      for (i = 0; i < M->n; i++) {
	M->data[j][i] *= scaling;
      }
    }
    return OE_SUCCESS;
  }
}
