/* oe_gaussian.h -- Gaussian smoothing for optimal estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <math.h>

#include "oe.h"

/* Reciprocal of square-root of 2 */
#define SQRT1_2 0.70710678118654752440

/* Calculate the integral of a standard-normal distribution from minus
   infinity to x. The result will lie between 0 and 1. Adapted from a
   function for erfc() in Numerical Recipies. */
static
OEreal
gaussian_integral(OEreal x)
{
  OEreal t, z, ans;
  z = fabs(x*SQRT1_2);
  t = 1.0 / (1.0 + 0.5*z);
  ans = 0.5*t * exp(- z*z
		    - 1.26551223
		    + t*(1.00002368 + 
			 t*(0.37409196 + 
			    t*(0.09678418 + 
			       t*(-0.18628806 +
				  t*(0.27886807 +
				     t*(-1.13520398 +
					t*(1.48851587 +
					   t*(-0.82215223 +
					      t*0.17087277)))))))));
  return x <= 0.0 ? ans : 1.0-ans;
}

/* Create a Gaussian smoothing matrix M that can be used to convert
   (usually high-resolution) data on range grid "range_n" to range
   grid "range_m" with Gaussian smoothing of standard deviation
   "sigma_range". The range grid "range_n" need not be evenly spaced;
   each element i is assumed to represent a top-hat function between
   range_n[i]-drange/2 and range_n[i]+drange/2. */
int
oe_gaussian_smoothing_matrix(OEmatrix *M, OEvector *range_m, OEvector *range_n,
			     OEreal drange, OEreal sigma_range)
{
  int m = range_m->n;
  int n = range_n->n;
  int i, j;
  OEreal drange_2 = 0.5*drange;
  OEreal norm = 1.0/sigma_range;

  OE_CHECK(oe_reset_matrix(M, m, n));
  
  for (i = 0; i < m; i++) {
    OEreal range = range_m->data[i];
    for (j = 0; j < n; j++) {
      OEreal sigmas1 = norm * (range - range_n->data[j] - drange_2);
      OEreal sigmas2 = norm * (range - range_n->data[j] + drange_2);

      if (sigmas2 <= -3.0) {
	continue;
      }
      else if (sigmas1 > 3.0) {
	continue;
	/*	break;*/
      }
      else if (sigmas1 <= -3.0 && sigmas2 > 3.0) {
	M->data[i][j] = 1.0;
	break;	
      }
      else if (sigmas1 <= -3.0 && sigmas2 <= 3.0) {
	M->data[i][j] = gaussian_integral(sigmas2);
      }
      else if (sigmas1 > -3.0 && sigmas2 <= 3.0) {
	M->data[i][j] = gaussian_integral(sigmas2)-gaussian_integral(sigmas1);
      }
      else if (sigmas1 > -3.0 && sigmas2 > 3.0) {
	M->data[i][j] = 1.0-gaussian_integral(sigmas1);
      }
    }
  }
  return OE_SUCCESS;
}
