/* oe_interp.h -- Interpolation functions for optimal estimation
   library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "oe.h"

/* Perform linear 1D interpolation: for each point xi[0..ni-1] find
   the value of y from the lookup table y(x) and return the result in
   yi_out. Also returned is the gradient of the function at xi,
   returned in dy_dx_out. Extrapolation is used if xi is out of range.
   This function is therefore suitable for use in forward models where
   x is a state variable, Z is the predicted observation and dZ_dx is
   the Jacobian. */
int
oe_1d_interp(OEvector *x, OEvector *y, int ni, OEreal *xi,
	     OEreal *yi_out, OEreal *dy_dx_out)
{
  int n = x->n, i;
  OEreal *xdata = x->data;
  OEreal *ydata = y->data;

  if (n != y->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Interpolation vectors must be the same length");
  }
  else if (n < 2) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Interpolation vectors must have at least 2 elements");
  }
  else if (xdata[0] >= xdata[n-1]) {
    OE_THROW(OE_MATRIX_ERROR,
	     "Interpolation x vector must be monotonically increasing");
  }

  for (i = 0; i < ni; i++, xi++, yi_out++) {
    int imin = 0, imax = n-1;

    if (*xi <= xdata[0]) {
      /* Extrapolate leftwards */
      imax = 1;
    }
    else if (*xi >= xdata[imax]) {
      /* Extrapolate rightwards */
      imin = imax-1;
    }
    else {
      int imid;
      /* Find pair in which xi sits */
      while (imax > imin+1) {
	imid = imin + (imax-imin)/2;
	if (*xi > xdata[imid]) {
	  imin = imid;
	}
	else {
	  imax = imid;
	}
      }
    }
    *yi_out = ((*xi-xdata[imin])*ydata[imax] + (xdata[imax]-*xi)*ydata[imin])
      / (xdata[imax]-xdata[imin]);

    if (dy_dx_out) {
      *dy_dx_out = (ydata[imax]-ydata[imin]) / (xdata[imax]-xdata[imin]);
      dy_dx_out++;
    }
  }

  return OE_SUCCESS;
}

/* Simple linear 2D interpolation: the value Zi_out is found for each
   point (xi, yi) from 0 to ni, using the lookup table Z(x, y). Also
   returned is the partial derivative with respect to x, which is
   returned in dZ_dx_out. As regards extrapolation, the x dimension is
   treated exactly the same as in the oe_1d_interp() function, but
   out-of-range yi values are clamped at the nearest valid y. This
   function is therefore suitable for use in forward models where x is
   a state variable, Z is the predicted observation, dZ_dx is the
   Jacobian, and y is a parameter that is assumed to be known
   accurately. */
int
oe_2d_interp(OEvector *x, OEvector *y, OEmatrix *Z,
	     int ni, OEreal *xi, OEreal *yi,
	     OEreal *Zi_out, OEreal *dZ_dx_out)
{
  int n = x->n, m = y->n, i;
  OEreal *xdata = x->data;
  OEreal *ydata = y->data;
  OEreal **Zdata = Z->data;

  if (n != Z->n || m != Z->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Interpolation vectors must match matrix dimensions");
  }
  else if (n < 2 || m < 2) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Interpolation vectors must have at least 2 elements");
  }
  else if (xdata[0] >= xdata[n-1] || ydata[0] >= ydata[m-1]) {
    OE_THROW(OE_MATRIX_ERROR,
	     "Interpolation vectors must be monotonically increasing");
  }

  for (i = 0; i < ni; i++, xi++, yi++, Zi_out++) {
    int imin = 0, imax = n-1;
    int jmin = 0, jmax = m-1;
    OEreal yi_mod = *yi;
    OEreal Zlowx, Zhighx;

    /* Find interpolation indeces in y direction */
    if (yi_mod <= ydata[0]) {
      /* Clamp value at lowest in lookup table */
      yi_mod = ydata[0];
      jmax = 1;
    }
    else if (yi_mod >= ydata[jmax]) {
      /* Clamp value at highest in lookup table */
      yi_mod = ydata[jmax];
      jmin = jmax-1;
    }
    else {
      int jmid;
      /* Find pair in which yi sits */
      while (jmax > jmin+1) {
	jmid = jmin + (jmax-jmin)/2;
	if (yi_mod > ydata[jmid]) {
	  jmin = jmid;
	}
	else {
	  jmax = jmid;
	}
      }
    }

    /* Find interpolation indeces in x direction */
    if (*xi <= xdata[0]) {
      /* Extrapolate leftwards */
      jmax = 1;
    }
    else if (*xi >= xdata[imax]) {
      /* Extrapolate rightwards */
      jmin = imax-1;
    }
    else {
      int imid;
      while (imax > imin+1) {
	imid = imin + (imax-imin)/2;
	if (*xi > xdata[imid]) {
	  imin = imid;
	}
	else {
	  imax = imid;
	}
      }
    }

    Zlowx = ((yi_mod-ydata[jmin])*Zdata[jmax][imin] 
	     + (ydata[jmax]-yi_mod)*Zdata[jmin][imin])
      / (ydata[jmax]-ydata[jmin]);
    Zhighx = ((yi_mod-ydata[jmin])*Zdata[jmax][imax] 
	      + (ydata[jmax]-yi_mod)*Zdata[jmin][imax])
      / (ydata[jmax]-ydata[jmin]);
    
    *Zi_out = ((*xi-xdata[imin])*Zhighx + (xdata[imax]-*xi)*Zlowx)
      / (xdata[imax]-xdata[imin]);

    if (dZ_dx_out) {
      *dZ_dx_out = (Zhighx-Zlowx) / (xdata[imax]-xdata[imin]);
      dZ_dx_out++;
    }
  }
  
  return OE_SUCCESS;
}
