/* oe_lu.h -- LU decomposition functions for optimal estimation
   library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <math.h>

#include "oe.h"

#define TINY 1.0e-20

/* Perform LU decomposition with pivotting on matrix A, returning the
   result in A and information on the interchanged rows in
   "index". This is adapted from Numerical Recipies. */
int
oe_lu_decomposition(OEmatrix *A, int *index, int *interchanges_even)
{
  static OEvector vect = OE_NEW_VECTOR;
  int i, imax = 0, j, k;
  OEreal big, dum, sum, temp;
  OEreal **A_data = A->data;
  int n = A->n;
  OEreal *vv;

  if (n != A->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Attempt to LU-decompose a non-square matrix");
  }

  OE_CHECK(oe_reset_vector(&vect, n));
  vv = vect.data;

  *interchanges_even = 1;
  for (i = 0; i < n; i++) {
    big = 0.0;
    for (j = 0; j < n; j++) {
      if ((temp = fabs(A_data[i][j])) > big) {
	big = temp;
      }
    }
    if (big == 0.0) {
      OE_THROW(OE_MATRIX_ERROR, "Attempt to LU-decompose a singular matrix");
    }
    vv[i] = 1.0/big;
  }
  for (j = 0; j < n; j++) {
    for (i = 0; i < j; i++) {
      sum = A_data[i][j];
      for (k = 0; k < i; k++) {
	sum -= A_data[i][k] * A_data[k][j];
      }
      A_data[i][j] = sum;
    }
    big = 0.0;
    for (i = j; i < n; i++) {
      sum = A_data[i][j];
      for (k = 0; k < j; k++) {
	sum -= A_data[i][k] * A_data[k][j];
      }
      A_data[i][j] = sum;
      if ((dum = vv[i]*fabs(sum)) >= big) {
	big = dum;
	imax = i;
      }
    }
    if (j != imax) {
      for (k = 0; k < n; k++) {
	dum = A_data[imax][k];
	A_data[imax][k] = A_data[j][k];
	A_data[j][k] = dum;
      }
      *interchanges_even = !(*interchanges_even);
      vv[imax] = vv[j];
    }
    index[j] = imax;
    if (A_data[j][j] == 0.0) {
      A_data[j][j] = TINY;
    }
    if (j != n) {
      dum = 1.0/(A_data[j][j]);
      for (i = j+1; i < n; i++) {
	A_data[i][j] *= dum;
      }
    }
  }
  return OE_SUCCESS;
}

/* Interpret "A" and "index" as the output of oe_lu_decomposition()
   and use them to solve the system of equations A*x=b (where here A
   is the matrix before its decomposition). The vector b is
   overwritten with the result x.*/
int
oe_lu_substitution(OEmatrix *A, int *index, OEvector *b)
{
  OEreal **A_data = A->data;
  OEreal *b_data = b->data;
  OEreal sum;
  int i, ii = -1, ip, j;
  int n = A->n;

  if (n != A->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Attempt to backsubstitute a non-square matrix");
  }
  
  if (n != b->n) {
    OE_THROW(OE_DIMENSION_ERROR,
     "Vector and matrix have incompatible dimensions in LU substitution");
  }

  for (i = 0; i < n; i++) {
    ip = index[i];
    sum = b_data[ip];
    b_data[ip] = b_data[i];
    if (ii >= 0) {
      for (j = ii; j < i; j++) {
	sum -= A_data[i][j] * b_data[j];
      }
    }
    else if (sum) {
      ii = i;
    }
    b_data[i] = sum;
  }
  for (i = n-1; i >= 0; i--) {
    sum = b_data[i];
    for (j = i+1; j < n; j++) {
      sum -= A_data[i][j] * b_data[j];
    }
    b_data[i] = sum/A_data[i][i];
  }
  return OE_SUCCESS;
}

