/* oe_spline.h -- Linear and cubic splines for optimal estimation
   library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include "oe.h"

/* Create a linear interpolation matrix M with dimensions m*n, that
   can be used to convert a vector of control points x to a high
   resolution vector of interpolated values x_hr using x_hr=M*x. */
int
oe_linear_interpolation_matrix(OEmatrix *M, int m, int n)
{
  if (n > m) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Meaningless interpolation matrix with n > m");
  }
  else if (m < 1 || n < 1) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Matrix dimensions must be greater than zero");
  }
  OE_CHECK(oe_reset_matrix(M, m, n));
  OE_CHECK(oe_zero_matrix(M));
  
  if (n > 1) {
    int i, j = 0;
    for (i = 0; i < n-1; i++) {
      int j1 = j;
      int j2 = (int) ((i+1)*(m-1))/(n-1);
      OEreal dj = j2-j1;
      for ( ; j < j2; j++) {
	M->data[j][i+1] = (1.0 - (M->data[j][i]=(j2-j)/dj));
      }
    }
    M->data[j][i] = 1.0;
  }
  else {
    int j;
    for (j = 0; j < m; j++) {
      M->data[j][0] = 1.0;
    }
  }
  return OE_SUCCESS;
}

/* Create a cubic basis spline interpolation matrix M with dimensions
   m*n, that can be used to convert a vector of control points x to a
   high resolution vector of interpolated values x_hr using
   x_hr=M*x. */
int
oe_cubic_b_spline(OEmatrix *M, int m, int n)
{
  if (n > m) {
    OE_THROW(OE_DIMENSION_ERROR, "Meaningless B-spline matrix with n > m");
  }
  else if (m < 1 || n < 1) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Matrix dimensions must be greater than zero");
  }
  OE_CHECK(oe_reset_matrix(M, m, n));
  OE_CHECK(oe_zero_matrix(M));
  
  if (n > 1) {
    int j;
    for (j = 0; j < m; j++) {
      OEreal i_j = ((OEreal) j*(n-1))/((OEreal) m-1);
      int ioffset = (int) i_j;
      OEreal u, u2, u3;
      u = i_j-ioffset;
      u2 = u*u;
      u3 = u*u*u;
      M->data[j][ioffset] = (2.0/3.0) - u2 + 0.5*u3;
      if (ioffset < n-1) {
	M->data[j][ioffset+1] = (1.0/6.0) + 0.5*u + 0.5*u2 - 0.5*u3;
      }
      else {
	M->data[j][ioffset] += (1.0/6.0) + 0.5*u + 0.5*u2 - 0.5*u3;
      }
      if (ioffset > 0) {
	M->data[j][ioffset-1] = (1-u)*(1-u)*(1-u)/6.0;
      }
      else {
	M->data[j][ioffset] += (1-u)*(1-u)*(1-u)/6.0;
      }
      if (ioffset < n-2) {
	M->data[j][ioffset+2] = u3/6.0;
      }
      /* Previously this line was just "else": memory overrun?  */
      else if (ioffset < n-1) {
	M->data[j][ioffset+1] += u3/6.0;
      }
    }
  }
  else {
    int j;
    for (j = 0; j < m; j++) {
      M->data[j][0] = 1.0;
    }
  }
  return OE_SUCCESS;
}
