/* 
   Copyright (C) 2000 Robin Hogan

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <new>
#include <fstream>
#include <iostream>
#include <cmath>
#include <cstdio>
#include <cstring>
#include "rplot.h"

#define PI_OVER_180 0.01745329251994329576

#define FIGURE_INTERP(x1,x2,x,w0,ww) (((double) (w0)) + \
		(((double) (ww))*((x)-(x1)))/((x2)-(x1)))
/* Changed the erroneous sign 24/07/00 */
#define OLD_FIGURE_PRETNI(x1,x2,w,w0,ww) ((x1) + \
		((x2)-(x1))*(((double) (w))+((double) (w0)))/((double) (ww)))
#define FIGURE_PRETNI(x1,x2,w,w0,ww) ((x1) + \
		((x2)-(x1))*(((double) (w))-((double) (w0)))/((double) (ww)))
#define ORDER_INT(a1,a2) if (a1>a2) {int tmp=a1;a1=a2;a2=tmp;}
#define ORDER_DOUBLE(a1,a2) if (a1>a2) {double tmp=a1;a1=a2;a2=tmp;}

// *** RFigure class ***

RFigure::RFigure(int width, int height, FORMAT format, ostream *os,
		 char *title, float ps_scale)
{
  int i;
  _sizeofheader=0;
  _width=width;
  _height=height;
  _xoffset=0;
  _yoffset=0;
  _format=format;
  _line_type=LINE_SOLID;
  _pen_color=make_color(0,0,0);
  if (_width < 1 || _height < 1) {
    throw ("Attempt to create a RFigure with a negative or zero width or height");
  }
  _clipbox=make_rectangle(0,0,1,1);
  axis(_clipbox);
  _clipping=false;
  _sizeofpixel = 0;

  switch(_format) {
  case BITMAP_8BIT:
    _sizeofpixel = 1;
    break;
  case BITMAP_24BIT:
    _sizeofpixel = 3;
    _sizeofheader=26;
    break;
  case BITMAP_32BIT:
    _sizeofpixel = 4;
    break;
  case POSTSCRIPT:
    if (os == NULL) {
      throw ("Attempt to create Postscript RFigure without specifying output stream");
    }
    else {
      _is_bitmap=false;
      _ps_stream=os;
      _init_postscript(title, ps_scale);
    }
    return;
  default:
    _sizeofpixel = 0;
    _is_bitmap=false;
    throw ("Attempt to create a RFigure with an unsupported format");
  }
  _is_bitmap=true;
  if (_sizeofpixel > 0) {
    int blocksize=_height * _width * _sizeofpixel;
    _pixels = new uchar[_sizeofheader + blocksize + (_height + 1) * 
		       sizeof(char *)];
    _data = new(_pixels +_sizeofheader + blocksize)	(uchar *);
    for (i=0 ; i<_height ; i++) {
      _data[i]=_pixels + _sizeofheader + i*_width*_sizeofpixel;
    }
    _data[_height]=NULL;
  }
  if (_format == BITMAP_24BIT && _sizeofheader > 0) {
    // set up PPMRAW header
    for (i=snprintf((char *) _pixels,_sizeofheader,"P6\n%d %d\n%d",width,
		    height,255);i<_sizeofheader-1;i++) {
      _pixels[i]=' ';
    }
    _pixels[_sizeofheader-1]='\n';
  }
}

RFigure::~RFigure(void)
{
  if (_is_bitmap) {
    delete[] _pixels;
  }
  else if (_format == POSTSCRIPT) {
    _close_postscript();
  }
}

void
RFigure::_rectangle_bitmap(int i1, int j1, int i2, int j2, const RColor &col)
{
  uchar value,r,g,b,a;
  int i,j,mini,minj,maxi,maxj;
  r=col.red;g=col.green;b=col.blue;

  ORDER_INT(i1,i2);
  ORDER_INT(j1,j2);
  mini=MAX(i1,0);
  maxi=MIN(i2,_width-1);
  // Reversing height in pixmap...
  minj=_height-MIN(j2,_height-1)-1;
  maxj=_height-MAX(j1,0)-1;

  if (mini>maxi || minj>maxj) {
    return;
  }

  switch (_format) {
  case BITMAP_8BIT:
    value=(palette+=col);
    for (j=minj;j<=maxj;j++) { for (i=mini;i<=maxi;i++) {
      _data[j][i]=value;
    } }
    break;
  case BITMAP_24BIT:
    for (j=minj;j<=maxj;j++) { for (i=mini;i<=maxi;i++) {
      _data[j][i*_sizeofpixel]=r;
      _data[j][i*_sizeofpixel+1]=g;
      _data[j][i*_sizeofpixel+2]=b;
    } }
    break;
  case BITMAP_32BIT:
    r=col.red;g=col.green;b=col.blue;a=col.alpha;
    for (j=minj;j<=maxj;j++) { for (i=mini;i<=maxi;i++) {
      _data[j][i*_sizeofpixel]=r;
      _data[j][i*_sizeofpixel+1]=g;
      _data[j][i*_sizeofpixel+2]=b;
      _data[j][i*_sizeofpixel+3]=a;
    } }
    break;
  }
}

void
RFigure::clear(const RColor &col)
{
  if (_is_bitmap) {
    _rectangle_bitmap(0,0,_width,_height,col);	
  }
  else if (_format == POSTSCRIPT) {
    _rectangle_postscript(0,0,_width,_height,col);
  }
}

void
RFigure::rectangle(double x1, double y1, double x2, double y2,
		   const RColor &col)
{
  double minx,maxx,miny,maxy;

  if (_clipping) {
    if (x1>x2) { minx=x2;maxx=x1; } else { minx=x1;maxx=x2; };
    if (y1>y2) { miny=y2;maxy=y1; } else { miny=y1;maxy=y2; };
    x1=MAX(minx,_clipbox.x1);
    x2=MIN(maxx,_clipbox.x2);
    y1=MAX(miny,_clipbox.y1);
    y2=MIN(maxy,_clipbox.y2);
    if (x1>x2 || y1>y2) {
      return;
    }
  }
  minx=FIGURE_INTERP(_axis.x1,_axis.x2,x1,_xoffset,_width);
  maxx=FIGURE_INTERP(_axis.x1,_axis.x2,x2,_xoffset,_width);
  miny=FIGURE_INTERP(_axis.y1,_axis.y2,y1,_yoffset,_height);
  maxy=FIGURE_INTERP(_axis.y1,_axis.y2,y2,_yoffset,_height);

  if (_is_bitmap) {
    RFigure::_rectangle_bitmap((int) minx, (int) miny, (int) maxx+1*RPLOT_FILL,
			       (int) maxy+1*RPLOT_FILL, col);
  }
  else if (_format == POSTSCRIPT) {
    _rectangle_postscript(minx,miny,maxx+0.1*RPLOT_FILL,maxy+0.1*RPLOT_FILL,col);
  }
}

//void pixel(int ix, int iy, const RColor &col);

void
RFigure::_pixel_bitmap(int i, int j, const RColor &col)
{
  int jr;
  if (i<0 || i>_width-1 || j<0 || j>_height-1) {
    return;
  }
  switch (_format) {
  case BITMAP_8BIT:
    _data[_height-1-j][i]=(palette+=col);
    return;
  case BITMAP_24BIT:
    jr=_height-1-j;
    _data[jr][i*_sizeofpixel]=col.red;
    _data[jr][i*_sizeofpixel+1]=col.green;
    _data[jr][i*_sizeofpixel+2]=col.blue;
    break;
  case BITMAP_32BIT:
    jr=_height-1-j;
    _data[jr][i*_sizeofpixel]=col.red;
    _data[jr][i*_sizeofpixel+1]=col.green;
    _data[jr][i*_sizeofpixel+2]=col.blue;
    _data[jr][i*_sizeofpixel+3]=col.alpha;
    break;
  }
}

void
RFigure::_plot_bitmap(int i1, int j1, int i2, int j2, const RColor &col,
		      LINE line_type)
{

  int i,j,mini,maxi,minj,maxj,jr,incr=1;
  uchar val=0,r=0,g=0,b=0,a=255;

  if (line_type == LINE_NONE) {
    return;
  }
  if (line_type == LINE_DOTTED) {
    incr=4;
  }
  if (i1==i2 && j1==j2) {
    _pixel_bitmap(i1,j1,col);
    return;
  }

  if (_clipping) {
    mini = MAX((int) FIGURE_INTERP(_axis.x1,_axis.x2,_clipbox.x1,_xoffset,_width),0);
    maxi = MIN((int) FIGURE_INTERP(_axis.x1,_axis.x2,_clipbox.x2,_xoffset,_width)+1,_width);
    minj = MAX((int) FIGURE_INTERP(_axis.y1,_axis.y2,_clipbox.y1,_yoffset,_height),0);
    maxj = MIN((int) FIGURE_INTERP(_axis.y1,_axis.y2,_clipbox.y2,_yoffset,_height)+1,_height);
  }
  else {
    mini=0;
    maxi=_width;
    minj=0;
    maxj=_height;
  }

  if ( (i1>maxi && i2>maxi) || (i1<mini && i2<mini)
       || (j1>maxj && j2>maxj) || (j1<minj && j2<minj) ) {
    return;
  }

  switch (_format) {
  case BITMAP_8BIT:
    val=(palette+=col);
    break;
  case BITMAP_24BIT:	case BITMAP_32BIT:
    r=col.red;g=col.green;b=col.blue;a=col.alpha;
    break;
  }
  if (ABS(i2-i1) > ABS(j2-j1)) {
    // 1 level of recursion to cope with wide brush:
    if (line_type == LINE_THICK) {
      _plot_bitmap(i1,j1+1,i2,j2+1,col,LINE_SOLID);
    }
    if (i2 < i1) {
      int tmp;
      tmp=i2;i2=i1;i1=tmp;
      tmp=j2;j2=j1;j1=tmp;
    }
    for (i=i1;i<=i2;i+=incr) {
      if (i>maxi-1) { return; };
      j=j1 + (int) (((float)(j2-j1)) * ((float)(i-i1))/((float)(i2-i1)));
      jr=_height-1-j;
      if (i<mini || j<minj || j>maxj-1) {continue; };
      switch (_format) {
      case BITMAP_8BIT:
	_data[jr][i]=val;
	break;
      case BITMAP_24BIT:
	_data[jr][i*_sizeofpixel]=r;
	_data[jr][i*_sizeofpixel+1]=g;
	_data[jr][i*_sizeofpixel+2]=b;
	break;
      case BITMAP_32BIT:
	_data[jr][i*_sizeofpixel]=r;
	_data[jr][i*_sizeofpixel+1]=g;
	_data[jr][i*_sizeofpixel+2]=b;
	_data[jr][i*_sizeofpixel+3]=a;
	break;
      }
    }
  }
  else {
    // 1 level of recursion to cope with wide brush:
    if (line_type==LINE_THICK) {
      _plot_bitmap(i1+1,j1,i2+1,j2,col,LINE_SOLID);
    }
    if (j2 < j1) {
      int tmp;
      tmp=i2;i2=i1;i1=tmp;
      tmp=j2;j2=j1;j1=tmp;
    }
    for (j=j1;j<=j2;j+=incr) {
      if (j>maxj-1) { return; };
      i=i1 + (int) (((float)(i2-i1)) * ((float)(j-j1))/((float)(j2-j1)));
      jr=_height-1-j;
      if (i<mini || i>maxi-1 || j<minj) {continue; };
      switch (_format) {
      case BITMAP_8BIT:
	_data[jr][i]=val;
	break;
      case BITMAP_24BIT:
	_data[jr][i*_sizeofpixel]=r;
	_data[jr][i*_sizeofpixel+1]=g;
	_data[jr][i*_sizeofpixel+2]=b;
	break;
      case BITMAP_32BIT:
	_data[jr][i*_sizeofpixel]=r;
	_data[jr][i*_sizeofpixel+1]=g;
	_data[jr][i*_sizeofpixel+2]=b;
	_data[jr][i*_sizeofpixel+3]=a;
	break;
      }
    }
  }
}

void
RFigure::pixel(double x, double y, const RColor &col)
{
  if (_clipping && !_clipbox.in_strict(x,y)) {
    return;
  }
  x=FIGURE_INTERP(_axis.x1,_axis.x2,x,_xoffset,_width);
  y=FIGURE_INTERP(_axis.y1,_axis.y2,y,_yoffset,_height);
  if (_is_bitmap) {
    _pixel_bitmap((int) x, (int) y, col);
  }
  else if (_format == POSTSCRIPT) {
    _rectangle_postscript(x,y,x+1,y+1,col);
  }
}

void
RFigure::plot(double x1, double y1, double x2, double y2, const RColor &col,
	      LINE line_type)
{
  x1=FIGURE_INTERP(_axis.x1,_axis.x2,x1,_xoffset,_width);
  y1=FIGURE_INTERP(_axis.y1,_axis.y2,y1,_yoffset,_height);
  x2=FIGURE_INTERP(_axis.x1,_axis.x2,x2,_xoffset,_width);
  y2=FIGURE_INTERP(_axis.y1,_axis.y2,y2,_yoffset,_height);
  if (_is_bitmap) {
    _plot_bitmap((int) x1, (int) y1, (int) x2, (int) y2, col, line_type);
  }
  else if (_format == POSTSCRIPT) {
    _plot_postscript(x1,y1,x2,y2,col,line_type);
  }
}

void RFigure::barb(double x, double y, double dx, double dy, const RColor &col,
		   LINE line_type, double *xreturn, double *yreturn) {
  x=FIGURE_INTERP(_axis.x1,_axis.x2,x,_xoffset,_width);
  y=FIGURE_INTERP(_axis.y1,_axis.y2,y,_yoffset,_height);
  if (_is_bitmap) {
    _plot_bitmap((int) x, (int) y, (int) (x+dx), (int) (y+dy), col, line_type);
  }
  else if (_format == POSTSCRIPT) {
    _plot_postscript(x,y,x+dx,y+dy,col,line_type);
  }

  if (xreturn != NULL) {
    (*xreturn) = FIGURE_PRETNI(_axis.x1,_axis.x2,x+dx,_xoffset,_width);
  }
  if (yreturn != NULL) {
    (*yreturn) = FIGURE_PRETNI(_axis.y1,_axis.y2,y+dy,_yoffset,_height);
  }
}



void
RFigure::write_ppmraw(ostream *os)
{
  if (_format != BITMAP_24BIT) {
    throw ("Only 24 bit figures can be saved as ppmraw");
  }
  os->write((const char*) _pixels,
	    _width * _height * _sizeofpixel + _sizeofheader);
}

void
RFigure::write_ppmraw(char * filename)
{
  ofstream file;
  file.open(filename);
  if (!file) {
    throw("Unable to open ppmraw file for writing");
  }
  write_ppmraw(&file);
  file.close();
}

RRectangle
RFigure::text(double x, double y, const char *str, ANCHOR anchor,
	      int rotation, RFont *fnt, const RColor &col)
{
  double swidth, sheight, width, height, xoffset=0;
  double yoffset=0, dx=0, dy=0;
  double rot, sin_rot, cos_rot, gap;
  RText text;

  if (fnt==NULL) {
    fnt=font;
  }
  gap=3; //fnt->_gap;
  x=FIGURE_INTERP(_axis.x1,_axis.x2,x,_xoffset,_width);
  y=FIGURE_INTERP(_axis.y1,_axis.y2,y,_yoffset,_height);

  if (rotation < 0) {
    rotation = 360-( (-rotation) % 360);
  }
  else {
    rotation %= 360;
  }

  if ( _format != POSTSCRIPT ) {
    rotation=90 * ((rotation+45)/90);
  }

  text.render(str, fnt);

  rot = rotation * PI_OVER_180;
  sin_rot = sin(rot);
  cos_rot = cos(rot);
  //	swidth = fnt->xextent(str);
  //	sheight = fnt->yextent(str);
  swidth = text.width();
  sheight = text.height();
  width = swidth * fabs(cos_rot) + sheight * fabs(sin_rot);
  height = swidth * fabs(sin_rot) + sheight * fabs(cos_rot);
  //	if (rotation > 90 && rotation <180)  

  switch (anchor) {
  case ANCHOR_CENTER:
    xoffset = -0.5 * swidth * cos_rot + 0.5 * sheight * sin_rot;
    yoffset = -0.5 * swidth * sin_rot - 0.5 * sheight * cos_rot;
    break;
  case ANCHOR_TOP:
    xoffset = -0.5 * swidth * cos_rot + (gap + sheight) * sin_rot;
    yoffset = -0.5 * swidth * sin_rot - (gap + sheight) * cos_rot;
    dx = (-gap + 0.5 * sheight) * sin_rot;
    dy = (-gap - 0.5 * sheight) * cos_rot;
    break;
  case ANCHOR_TOPRIGHT:
    xoffset = -(gap + swidth) * cos_rot + (gap + sheight) * sin_rot;
    yoffset = -(gap + swidth) * sin_rot - (gap + sheight) * cos_rot;
    dx = (-gap - 0.5 * swidth) * cos_rot + (-gap + 0.5 * sheight) * sin_rot;
    dy = (-gap - 0.5 * swidth) * sin_rot - (gap + 0.5 * sheight) * cos_rot;
    break;
  case ANCHOR_RIGHT:
    xoffset = -(gap + swidth) * cos_rot + 0.5 * sheight * sin_rot;
    yoffset = -(gap + swidth) * sin_rot - 0.5 * sheight * cos_rot;
    dx = (-gap - 0.5 * swidth) * cos_rot;
    dy = (-gap - 0.5 * swidth) * sin_rot;
    break;
  case ANCHOR_BOTTOMRIGHT:
    xoffset = -(gap + swidth) * cos_rot - gap * sin_rot;
    yoffset = -(gap + swidth) * sin_rot + gap * cos_rot;
    dx = (-gap - 0.5 * swidth) * cos_rot + (gap + 0.5 * sheight) * sin_rot;
    dy = (-gap - 0.5 * swidth) * sin_rot + (gap - 0.5 * sheight) * cos_rot;
    break;
  case ANCHOR_BOTTOM:
    xoffset = -0.5 * swidth * cos_rot - gap * sin_rot;
    yoffset = -0.5 * swidth * sin_rot + gap * cos_rot;
    dx = (gap - 0.5 * sheight) * sin_rot;
    dy = (gap + 0.5 * sheight) * cos_rot;
    break;
  case ANCHOR_BOTTOMLEFT:
    xoffset = gap * cos_rot - gap * sin_rot;
    yoffset = gap * sin_rot + gap * cos_rot;
    dx = (gap + 0.5 * swidth) * cos_rot + (gap + 0.5 * sheight) * sin_rot;
    dy = (gap + 0.5 * swidth) * sin_rot + (gap - 0.5 * sheight) * cos_rot;
    break;
  case ANCHOR_LEFT:
    xoffset = gap * cos_rot + 0.5 * sheight * sin_rot;
    yoffset = gap * sin_rot - 0.5 * sheight * cos_rot;
    dx = (gap + 0.5 * swidth) * cos_rot;
    dy = (gap + 0.5 * swidth) * sin_rot;
    break;
  case ANCHOR_TOPLEFT:
    xoffset = gap * cos_rot + (gap + sheight) * sin_rot;
    yoffset = gap * sin_rot - (gap + sheight) * cos_rot;
    dx = (gap + 0.5 * swidth) * cos_rot + (-gap - 0.5 * sheight) * sin_rot;
    dy = (gap + 0.5 * swidth) * sin_rot + (-gap + 0.5 * sheight) * cos_rot;
    break;
  }

  if (_is_bitmap) {
    //		_text_bitmap(x+xoffset,y+yoffset,str,rotation,fnt,col);
    _text_bitmap((int) (x+xoffset),(int) (y+yoffset),&text,rotation,col);
  }
  else if (_format == POSTSCRIPT) {
    //		_text_postscript(x+xoffset,y+yoffset,str,rotation,fnt,ol);
    _text_postscript(x+xoffset,y+yoffset,&text,rotation,col);
  }
  return make_rectangle(
			FIGURE_PRETNI(_axis.x1,_axis.x2,x+dx-0.5*width,_xoffset,_width),
			FIGURE_PRETNI(_axis.y1,_axis.y2,y+dy-0.5*height,_yoffset,_height),
			FIGURE_PRETNI(_axis.x1,_axis.x2,x+dx+0.5*width,_xoffset,_width),
			FIGURE_PRETNI(_axis.y1,_axis.y2,y+dy+0.5*height,_yoffset,_height)
			);
}

void
RFigure::_text_bitmap(int x, int y, RText *text, int rotation,
		      const RColor &col)
{
  uword ch=0,r,b,pos=0;
  int ichar, char_width, char_height, offset,bpr, dr, db;
  uchar *row;
  char irot=rotation/90;
  list<RTextElement>::const_iterator i=text->element.begin();

  for ( ; i != text->element.end(); i++) {
    for (ch=0; ch<(i->text.size()); ch++) {
      ichar = (i->text[ch] < 0 ? (int) 256 - (i->text[ch]) : i->text[ch]);
      if (ichar < i->font->first || ichar > i->font->last) {
	ichar=i->font->def;
      }
      if (!(i->font->exists[ichar-(i->font->first)])) {
	ichar=i->font->def;
      }
      ichar -= i->font->first;
      char_width = i->font->right[ichar]-i->font->left[ichar];
      char_height = i->font->ascent[ichar]+i->font->descent[ichar];
      offset = i->font->goffset[ichar];
      bpr = ((char_width+7)>>3);
      for (r=0; r<char_height; r++ ) {
	row = i->font->glyphs + offset;
	dr=i->font->max_descent + i->font->ascent[ichar]-r + i->ascent;
	for (b=0; b<char_width; b++) {
	  if (row[b >> 3] & (1 << (7 - (b & 7)))) {
	    db=pos+b+i->font->left[ichar];
	    switch (irot) {
	    case 0:
	      _pixel_bitmap(x+db,y+dr,col);
	      break;
	    case 1:
	      _pixel_bitmap(x-dr,y+db,col);
	      break;
	    case 2:
	      _pixel_bitmap(x-db,y-dr,col);
	      break;
	    case 3:
	      _pixel_bitmap(x+dr,y-db,col);
	      break;
	    }
	  }
	}
	offset += bpr;
      }
      pos += i->font->width[ichar];
    }

  }
}

void
RFigure::polygon(int n, float *xx, float *yy, const RColor &col)
{
  float x[RPLOT_MAX_POLYGON_VERTICES+1], y[RPLOT_MAX_POLYGON_VERTICES+1];

  if (n > RPLOT_MAX_POLYGON_VERTICES) {
    throw("Too many vertices requested for polygon");
  }

  /* Get into raw coordinates and offset any integer values */
  for (register int i = 0; i < n; ++i) {
    x[i] = FIGURE_INTERP(_axis.x1,_axis.x2,xx[i],_xoffset,_width);
    y[i] = FIGURE_INTERP(_axis.y1,_axis.y2,yy[i],_yoffset,_height);
    if (x[i] == (float) ((int) x[i])) 
      x[i] += 0.0001;
    if (y[i] == (float) ((int) y[i])) 
      y[i] += 0.0001;
  }
  x[n] = x[0];
  y[n] = y[0];

  if (_is_bitmap) {
    _polygon_bitmap(n, x, y, col);
  }
  else if (_format == POSTSCRIPT) {
    _polygon_postscript(n, x, y, col);
  }
}

/* Simple scanline algorithm - OK for convex polygons */
void
RFigure::_polygon_bitmap(int n, float *x, float *y, const RColor &col)
{
  int i, iy, ix[2], in = 0, y1, y2, yr;
  int minx, maxx, miny, maxy;
  uchar val=0,r=0,g=0,b=0,a=255;

  //  cerr << xx[0] << " " << yy[0] << "; ";
  //  pixel(xx[0], yy[0], col);

  //  x = new float[2*n + 2]; // Too slow?
  //  y = x + (n + 1);

  y1 = (int) y[0] + 1;
  y2 = (int) y[0];
  /* Find extreme y values */
  for (i = 1; i < n; ++i) {
    if (y1 > y[i] + 1) {
      y1 = (int) y[i] + 1;
    }
    else if (y2 < y[i]) {
      y2 = (int) y[i];
    }
  }

  if (_clipping) {
    minx = MAX((int) FIGURE_INTERP(_axis.x1,_axis.x2,_clipbox.x1,_xoffset,_width),0);
    maxx = MIN((int) FIGURE_INTERP(_axis.x1,_axis.x2,_clipbox.x2,_xoffset,_width),_width);
    miny = MAX((int) FIGURE_INTERP(_axis.y1,_axis.y2,_clipbox.y1,_yoffset,_height),0);
    maxy = MIN((int) FIGURE_INTERP(_axis.y1,_axis.y2,_clipbox.y2,_yoffset,_height),_height);
  }
  else {
    minx = 0;
    maxx = _width;
    miny = 0;
    maxy = _height;
  }

  /* Is region within clip box? */
  y1 = MAX(y1, miny);
  y2 = MIN(y2, maxy);

  switch (_format) {
  case BITMAP_8BIT:
    val=(palette+=col);
    break;
  case BITMAP_24BIT:	case BITMAP_32BIT:
    r=col.red;g=col.green;b=col.blue;a=col.alpha;
    break;
  }

  if (y2 >= y1) {
    for (iy = y1; iy <= y2; ++iy) {
      /* Find crossing points */
      in = 0;
      for (i = 0; i < n; ++i) {
	if ((y[i] > iy && y[i+1] < iy) || (y[i] < iy && y[i+1] > iy)) {
	  ix[in++] = (int) (x[i] + (iy-y[i]) * (x[i+1]-x[i]) / (y[i+1]-y[i]));
	  if (in == 2) {
	    break;
	  }
	}
      }
      if (in < 2) {
	continue;
      }
      if (ix[0] > ix[1]) {
	register int tmp_ix = ix[1];
	ix[1] = ix[0];
	ix[0] = tmp_ix;
      }
      ix[0] = MAX(ix[0], minx);
      ix[1] = MIN(ix[1], maxx);
      yr = _height - 1 - iy;
      switch (_format) {
      case BITMAP_8BIT:
	for (int ii = ix[0]; ii < ix[1]; ++ii) {
	  _data[yr][ii]=val;
	}
	break;
      case BITMAP_24BIT:
	for (int ii = ix[0]; ii < ix[1]; ++ii) {
	  _data[yr][ii*_sizeofpixel]=r;
	  _data[yr][ii*_sizeofpixel+1]=g;
	  _data[yr][ii*_sizeofpixel+2]=b;
	}
	break;
      case BITMAP_32BIT:
	for (int ii = ix[0]; ii < ix[1]; ++ii) {
	  _data[yr][ii*_sizeofpixel]=r;
	  _data[yr][ii*_sizeofpixel+1]=g;
	  _data[yr][ii*_sizeofpixel+2]=b;
	  _data[yr][ii*_sizeofpixel+3]=a;
	}
	break;
      }
    }
  }
  // delete[] x;
}
