/* 
   Copyright (C) 2000 Robin Hogan

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <iostream>
#include <string>
#include <cctype>

#include "rplot.h"

bool
match(const string &a, char *sb)
{
  unsigned int i;
  string b=sb;
  if (a.length() != b.length()) {
    return false;
  }
  for (i=0;i<a.length();i++) {
    if (tolower(a[i]) != tolower(b[i])) {
      return false;
    }
  }
  return true;
}

ostream&
operator<<(ostream& os, const RTextElement &el)
{
  return (os << "<" << el.font->foundry << "/" << el.font->family 
	  << "/" << el.font->weight << "/" << el.font->slant
	  << "/" << el.font->pixel << "> " <<  el.text);
}

ostream&
operator<<(ostream& os, const RText &text)
{
  list<RTextElement>::const_iterator i=text.element.begin();
  os << "Size of list is " << text.element.size() << endl;
  for ( ; i != text.element.end(); i++) {
    os << *i << "\n";
  }
  return (os);
};

unsigned int
RTextElement::width() const
{
  int ichar;
  unsigned int ch, extent=0;
  for (ch=0; ch<text.size(); ch++) {
    ichar = (text[ch] < 0 ? (int) 256-text[ch] : text[ch]);
    if (ichar < font->first || ichar > font->last) {
      ichar=font->def;
    }
    if (!font->exists[ichar-font->first]) {
      ichar=font->def;
    }
    extent += font->width[ichar-font->first];
  }
  return extent;
};

unsigned int
RTextElement::height() const
{
  return font->max_ascent + font->max_descent;
}

unsigned int
RText::width()
{
  list<RTextElement>::const_iterator i=element.begin();
  unsigned int extent=0;
  for ( ; i != element.end(); i++) {
    extent += i->width();
  }
  return extent;
};

unsigned int
RText::height()
{
  int ascent=0, descent=0;
  list<RTextElement>::const_iterator i=element.begin();
  for ( ; i != element.end(); i++) {
    ascent = MAX(ascent, i->font->max_ascent + i->ascent);
    descent = MAX(descent, i->font->max_descent - i->ascent);
  }
  return (ascent + descent);
}

void
RText::render(const char *str, RFont *font)
{
  RFont *last_font=font, *current_font=font, *tmp_font;
  bool bold=match(font->weight, "bold"),
    italic=(match(font->slant, "i") || match(font->slant, "o"));
  string::size_type ch=0, next=0;
  bool not_finished=1;
  string remaining, current;
  int ascent=0;

  element.erase(element.begin(),element.end());
  while (not_finished) {
    remaining = str+ch;
    if ((next = remaining.find_first_of("<")) == string::npos) {
      current.append(remaining);
      element.push_back(make_text_element(current,current_font,ascent));
      not_finished=0;
      break;
    }
    else {
      if (remaining[next+1] == '<') {
	current.append(remaining.substr(0,next+1));
	element.push_back(make_text_element(current,current_font,ascent));
	current.erase();
	ch += next+2;
      }
      else {
	current.append(remaining.substr(0,next));
	element.push_back(make_text_element(current,current_font,ascent));
	current.erase();
	ch += next+1;

	remaining = str+ch;
	if ((next = remaining.find_first_of(">")) == string::npos) {
	  not_finished=0;
	  break;
	}
	else {
	  current.append(remaining.substr(0,next));
	  if (match(current,"sub")) {
	    ascent -= (font->max_ascent / 2);
	    if (font->pixel == current_font->pixel) {
	      int new_size = (int) 0.7*current_font->pixel;
	      if (new_size < 8) {
		new_size = MIN(current_font->pixel, 8);
	      }
	      if ((tmp_font=font->parent->load(new_size, current_font->family,
					       current_font->weight, current_font->slant,
					       current_font->foundry)) != NULL) {
		current_font = tmp_font;
	      };
	    }
	  }
	  else if (match(current,"/sub")) {
	    ascent += (font->max_ascent / 2);
	    if (font->pixel < current_font->pixel) {
	      if ((tmp_font 
		   = font->parent->load(font->pixel,
					current_font->family,
					current_font->weight,
					current_font->slant,
					current_font->foundry)) != NULL) {
		current_font = tmp_font;
	      };							
	    }
	  }
	  else if (match(current,"sup")) {
	    ascent += (font->max_ascent / 2);
	    if (font->pixel == current_font->pixel) {
	      int new_size = (int) 0.7*current_font->pixel;
	      if (new_size < 8) {
		new_size = MIN(current_font->pixel, 8);
	      }
	      if ((tmp_font
		   = font->parent->load(new_size,
					current_font->family,
					current_font->weight,
					current_font->slant,
					current_font->foundry)) != NULL) {
		current_font = tmp_font;
	      };							
	    }
	  }
	  else if (match(current,"/sup")) {
	    ascent -= (font->max_ascent / 2);
	    if (font->pixel > current_font->pixel) {
	      if ((tmp_font
		   = font->parent->load(font->pixel,
					current_font->family,
					current_font->weight,
					current_font->slant,
					current_font->foundry)) != NULL) {
		current_font = tmp_font;
	      };							
	    }
	  }
	  else if (match(current,"i")) {
	    if (!italic) {
	      if ((tmp_font 
		   = font->parent->load(current_font->pixel,
					current_font->family,
					current_font->weight,
					string("i"),
					current_font->foundry)) != NULL) {
		current_font = tmp_font; italic = true;
	      }
	      else {
		if ((tmp_font
		     = font->parent->load(current_font->pixel,
					  current_font->family,
					  current_font->weight,
					  string("o"),
					  current_font->foundry)) != NULL) {
		  current_font = tmp_font; italic= true;
		}
	      }
	    }
	  }
	  else if (match(current,"/i")) {
	    if (italic) {
	      if ((tmp_font 
		   = font->parent->load(current_font->pixel,
					current_font->family,
					current_font->weight,
					string("r"),
					current_font->foundry)) != NULL) {
		current_font = tmp_font; italic = false;
	      }
	    }
	  }
	  else if (match(current,"b")) {
	    if (!bold) {
	      if ((tmp_font
		   = font->parent->load(current_font->pixel,
					current_font->family,
					string("bold"),
					current_font->slant,
					current_font->foundry)) != NULL) {
		current_font = tmp_font; bold = true;
	      }
	    }
	  }
	  else if (match(current,"/b")) {
	    if (bold) {
	      if ((tmp_font
		   = font->parent->load(current_font->pixel,
					current_font->family,
					string("medium"),
					current_font->slant,
					current_font->foundry)) != NULL) {
		current_font = tmp_font; bold = false;
	      }
	    }
	  }
	  else if (match(current,"sym")) {
	    if ((tmp_font
		 = font->parent->load(current_font->pixel,
				      string("symbol"),
				      string("medium"),
				      string("r"),
				      string("adobe"))) != NULL) {
	      last_font = current_font; current_font = tmp_font;
	    }
	  }
	  else if (match(current,"/sym")) {
	    if ((tmp_font
		 = font->parent->load(current_font->pixel,
				      last_font->family,
				      last_font->weight,
				      last_font->slant,
				      last_font->foundry)) != NULL) {
	      current_font = tmp_font; last_font = font;
	    }
	  }
	  else {
	    cerr << "Warning: use of unidentified markup `<" << current << ">'\n";
	  }
	  current.erase();
	  ch += next+1;
	}
      }
    }
  }
}
