#include <string.h>
#include "oe_netcdf.h"
#include "nctools.h"
#include "zdr_phidp.h"

//#define NC_REAL NC_DOUBLE
#define NC_REAL NC_FLOAT

static
int
add_string_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    /* The +1 ensures that \0 is included at the end of each attribute
       in the netcdf file - is this necessary? */
    NC_CHECK(nc_put_att_text(ncid, fieldid, attname,
			     strlen(value), value));
  }
  return OE_SUCCESS;
}

static
int
add_attributes(int ncid, int fieldid, char *units,
               char *long_name, char *error_variable)
{
  if (units) {
    OE_CHECK(add_string_attribute(ncid, fieldid, "units", units));
  }
  if (long_name) {
    OE_CHECK(add_string_attribute(ncid, fieldid, "long_name", long_name));
  }
  if (error_variable) {
    OE_CHECK(add_string_attribute(ncid, fieldid, "error_variable",
				  error_variable));
  }
  return OE_SUCCESS;
}


static
int
add_plot_attributes(int ncid, int fieldid, char *units_html,
		    float plot_min, float plot_max, char is_log)
{
  float plot_range[2] = {plot_min, plot_max};
  if (units_html) {
    OE_CHECK(add_string_attribute(ncid, fieldid, "units_html", units_html));
  }
  NC_CHECK(nc_put_att_float(ncid, fieldid, "plot_range", NC_FLOAT, 
			    2, plot_range));
  if (is_log) {
    OE_CHECK(add_string_attribute(ncid, fieldid, "plot_scale",
				  "logarithmic"));
  }
  return OE_SUCCESS;
}

static
int
add_missing_value(int ncid, int fieldid, float value)
{
  NC_CHECK(nc_put_att_float(ncid, fieldid,
			    "missing_value",
			    NC_REAL, 1, &value));
  NC_CHECK(nc_put_att_float(ncid, fieldid,
			    "_FillValue",
			    NC_REAL, 1, &value));
  return OE_SUCCESS;
}

static
int
put_matrix(int ncid, int fieldid, OEmatrix *M)
{
  size_t start[2], count[2];
  int i;

  start[1] = 0;
  count[0] = 1;
  count[1] = M->n;

  for (i = 0; i < M->m; i++) {
    start[0] = i;
    NC_CHECK(nc_put_vara_double(ncid, fieldid, start, count, M->data[i]));
  }
  return OE_SUCCESS;
}

static
int
put_vector(int ncid, int fieldid, OEvector *v)
{
  size_t start[1], count[1];

  start[0] = 0;
  count[0] = v->n;

  NC_CHECK(nc_put_vara_double(ncid, fieldid, start, count, v->data));
  return OE_SUCCESS;
}

int
save_retrieval(char *filename, Retrieval *retrieval, 
	       Scan *scan, Settings *settings)
{
  /* IDs etc of the output file */
  int ncid;
  int rayid, gateid;
  int R_id, lnR_error_id, Ah_id, Av_id, a_id, lna_error_id;
  int Zdr_fwd_id, phi_dp_fwd_id;
  int as_id, as_error_id, hail_fraction_id, hail_fraction_error_id;
  int time_id, range_id, azim_id, elev_id;
  int dimids[2];

  /* IDs of the input file */
  int ncid_input, time_id_input, range_id_input;
  int elev_id_input, azim_id_input;

  /* Open original scan to get attributes from it */
  NC_CHECK(nc_open(settings->input, NC_NOWRITE, &ncid_input));
  NC_CHECK(nc_inq_varid(ncid_input, "range", &range_id_input));
  NC_CHECK(nc_inq_varid(ncid_input, "time", &time_id_input));
  NC_CHECK(nc_inq_varid(ncid_input, "azim", &azim_id_input));
  NC_CHECK(nc_inq_varid(ncid_input, "elev", &elev_id_input));

  /* Create retrieval file */
  NC_CHECK(nc_create(filename, NC_CLOBBER, &ncid));

  /* Define dimensions*/
  NC_CHECK(nc_def_dim(ncid, "time", retrieval->R.m, &rayid));
  NC_CHECK(nc_def_dim(ncid, "range", retrieval->R.n, &gateid));
  
  dimids[0] = rayid;
  dimids[1] = gateid;
  /* Define vectors */
  NC_CHECK(nc_def_var(ncid, "range", NC_FLOAT, 1, dimids+1, &range_id));
  NC_CHECK(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids, &time_id));
  NC_CHECK(nc_def_var(ncid, "azim", NC_FLOAT, 1, dimids, &azim_id));
  NC_CHECK(nc_def_var(ncid, "elev", NC_FLOAT, 1, dimids, &elev_id));

  /* Define matrices */
  NC_CHECK(nc_def_var(ncid, "R", NC_REAL, 2, dimids, &R_id));
  NC_CHECK(nc_def_var(ncid, "lnR_error", NC_REAL, 2, dimids, &lnR_error_id));
  NC_CHECK(nc_def_var(ncid, "a", NC_REAL, 2, dimids, &a_id));
  NC_CHECK(nc_def_var(ncid, "lna_error", NC_REAL, 2, dimids, &lna_error_id));
  NC_CHECK(nc_def_var(ncid, "Ah", NC_REAL, 2, dimids, &Ah_id));
  NC_CHECK(nc_def_var(ncid, "Av", NC_REAL, 2, dimids, &Av_id));
  NC_CHECK(nc_def_var(ncid, "Zdr_fwd", NC_REAL, 2, dimids, &Zdr_fwd_id));
  NC_CHECK(nc_def_var(ncid, "phi_dp_fwd", NC_REAL, 2, dimids, &phi_dp_fwd_id));
  if (settings->do_hail_retrieval) {
    NC_CHECK(nc_def_var(ncid, "hail_fraction", NC_REAL,
			2, dimids, &hail_fraction_id));
    NC_CHECK(nc_def_var(ncid, "hail_fraction_error", NC_REAL,
			2, dimids, &hail_fraction_error_id));
  }
  if (settings->var_atten_scaling > 0.0) {
    NC_CHECK(nc_def_var(ncid, "atten_scaling", NC_REAL, 1, dimids, &as_id));
    NC_CHECK(nc_def_var(ncid, "atten_scaling_error",
			NC_REAL, 1, dimids, &as_error_id));
  }

  /* Add attributes to vectors */
  NC_CHECK(nct_copy_attributes(ncid_input, range_id_input, ncid, range_id));
  NC_CHECK(nct_copy_attributes(ncid_input, time_id_input, ncid, time_id));
  NC_CHECK(nct_copy_attributes(ncid_input, elev_id_input, ncid, elev_id));
  NC_CHECK(nct_copy_attributes(ncid_input, azim_id_input, ncid, azim_id));

  /* Add attributes to matrices */
  OE_CHECK(add_attributes(ncid, R_id, "mm h-1", "Rainfall rate", "lnR_error"));
  OE_CHECK(add_plot_attributes(ncid, R_id, "mm h<sup>-1</sup>", 0.1, 100, 1));

  OE_CHECK(add_attributes(ncid, lnR_error_id, "1", 
    "1-sigma error in natural log of rainfall rate",
			  NULL));
  OE_CHECK(add_plot_attributes(ncid, lnR_error_id, NULL, 0, 1, 0));

  OE_CHECK(add_attributes(ncid, a_id, "(mm6 m-3) (mm h-1)^(1/1.5)",
			  "Coefficient a in Zh=a*R^b", "lna_error"));
  OE_CHECK(add_plot_attributes(ncid, a_id, 
	       "mm<sup>6</sup> m<sup>-3</sup> (mm h<sup>-1</sup>)<sup>1/1.5</sup>",
			       10, 1000, 1));

  OE_CHECK(add_attributes(ncid, lna_error_id, "1",
  "1-sigma error in natural log of coefficient a",
			  NULL));
  OE_CHECK(add_plot_attributes(ncid, lna_error_id, NULL, 0, 1, 0));

  OE_CHECK(add_attributes(ncid, Ah_id, "dB",
	  "Total 2-way attenuation, horizontal polarisation", NULL));
  OE_CHECK(add_plot_attributes(ncid, Ah_id, "dB", 0, 10, 0));

  OE_CHECK(add_attributes(ncid, Av_id, "dB",
	  "Total 2-way attenuation, vertical polarisation", NULL));
  OE_CHECK(add_plot_attributes(ncid, Av_id, "dB", 0, 10, 0));

  OE_CHECK(add_attributes(ncid, Zdr_fwd_id, "dB",
		  "Forward modelled differential reflectivity", NULL));
  OE_CHECK(add_plot_attributes(ncid, Zdr_fwd_id, "dB", -1, 4, 0));

  OE_CHECK(add_attributes(ncid, phi_dp_fwd_id, "degrees",
		  "Forward modelled 2-way differential phase shift", NULL));
  OE_CHECK(add_plot_attributes(ncid, phi_dp_fwd_id, NULL, -10, 50, 0));

  if (settings->do_hail_retrieval) {
    OE_CHECK(add_attributes(ncid, hail_fraction_id, "1",
    "Fraction of unattenuated radar reflectivity due to hail",
			    "hail_fraction_error"));
    OE_CHECK(add_plot_attributes(ncid, hail_fraction_id, NULL, 0, 1, 0));

    OE_CHECK(add_attributes(ncid, hail_fraction_error_id, "1",
	    "1-sigma error in hail reflectivity fraction", NULL));
    OE_CHECK(add_plot_attributes(ncid, hail_fraction_error_id, NULL, 0, 1, 0));
  }
  if (settings->var_atten_scaling > 0.0) {
    OE_CHECK(add_attributes(ncid, as_id, "1",
			    "Radar attenuation scaling factor", 
			    "atten_scaling_error"));
    OE_CHECK(add_plot_attributes(ncid, as_id, NULL, 0, 3, 0));

    OE_CHECK(add_attributes(ncid, as_error_id, "1",
    "Error in radar attenuation scaling factor, 1 standard deviation",
			    NULL));
    OE_CHECK(add_plot_attributes(ncid, as_error_id, NULL, 0, 1, 0));
  }

  /* Global attributes */
  NC_CHECK(nct_copy_attributes(ncid_input, NC_GLOBAL, ncid, NC_GLOBAL));
  NC_CHECK(nct_add_history(ncid, "Variational retrieval run", NULL));
  NC_CHECK(nct_add_command_line(ncid, settings->argc, settings->argv));
  NC_CHECK(add_string_attribute(ncid, NC_GLOBAL, "title",
				"Output of variational retrieval"))
  NC_CHECK(add_string_attribute(ncid, NC_GLOBAL, "references",
	"Hogan, R. J., 2007: J. Appl. Meteorol. Climatology, 46, 1544-1564"));
  nc_del_att(ncid, NC_GLOBAL, "software_version");
  NC_CHECK(nc_put_att_text(ncid, NC_GLOBAL, "config",
			   strlen(settings->confstring),
			   settings->confstring));

  /* Now write the numerical data */
  NC_CHECK(nc_close(ncid_input));
  NC_CHECK(nc_enddef(ncid));

  /* Write data to vectors */
  OE_CHECK(put_vector(ncid, range_id, &scan->range));
  OE_CHECK(put_vector(ncid, time_id, &scan->time));
  OE_CHECK(put_vector(ncid, azim_id, &scan->azim));
  OE_CHECK(put_vector(ncid, elev_id, &scan->elev));

  /* Write data to matrices */
  OE_CHECK(put_matrix(ncid, R_id, &retrieval->R));
  OE_CHECK(put_matrix(ncid, lnR_error_id, &retrieval->lnR_error));
  OE_CHECK(put_matrix(ncid, a_id, &retrieval->a));
  OE_CHECK(put_matrix(ncid, lna_error_id, &retrieval->lna_error));
  OE_CHECK(put_matrix(ncid, Ah_id, &retrieval->Ah));
  OE_CHECK(put_matrix(ncid, Av_id, &retrieval->Av));
  OE_CHECK(put_matrix(ncid, Zdr_fwd_id, &retrieval->Zdr_fwd));
  OE_CHECK(put_matrix(ncid, phi_dp_fwd_id, &retrieval->phi_dp_fwd));
  if (settings->do_hail_retrieval) {
    OE_CHECK(put_matrix(ncid, hail_fraction_id, &retrieval->hail_fraction));
    OE_CHECK(put_matrix(ncid, hail_fraction_error_id,
			&retrieval->hail_fraction_error));
  }
  if (settings->var_atten_scaling > 0.0) {
    OE_CHECK(put_vector(ncid, as_id, &retrieval->atten_scaling));
    OE_CHECK(put_vector(ncid, as_error_id, &retrieval->atten_scaling_error));
  }

  NC_CHECK(nc_close(ncid));
  return OE_SUCCESS;
}
