#include <string.h>
#include "oe.h"

typedef struct {
  OEmatrix Zh;
  OEmatrix Zv;
  OEmatrix attenh;
  OEmatrix attenv;
  OEmatrix Kdp;
  OEvector diameter;
  OEvector temperature;
  OEvector K2;
} ScatteringData;

typedef struct {
  OEmatrix ln_Zh_R;
  OEmatrix ln_Zh_Kdp;
  OEmatrix ln_Zh_attenh;
  OEmatrix ln_Zh_attenv;
  OEmatrix Zdr;
  OEvector temperature;
  OEvector D0;
  OEreal mu;
  OEreal hail_Zdr_lin;
  OEreal hail_Kdp_Zh;
  OEreal hail_Ah_Zh;
  OEreal hail_Av_Zh;
} ScatteringLUT;

typedef struct {
  OEmatrix Zh;
  OEmatrix Zdr;
  OEmatrix phi_dp;
  OEmatrix Zh_error;
  OEmatrix Zdr_error;
  OEmatrix phi_dp_error;
  OEmatrix temperature;
  OEvector range;
  OEvector azim;
  OEvector elev;
  OEvector time;
  OEreal frequency;
  int is_Zdr;
  int is_phi_dp;
  int is_Zh_error;
} Scan;

typedef struct {
  OEreal a_default;
  OEreal b;
  OEreal radial_decorr_dist;
  OEreal azim_decorr_dist;
  OEreal var_ln_a_apriori;
  OEreal max_ln_a_excursion;
  OEreal Zdr_error_factor;
  OEreal phi_dp_error_factor;
  OEreal atten_scaling;
  OEreal max_atten;
  OEreal var_atten_scaling;
  OEreal atten_scaling_decorr;
  OEreal min_hail_Zh;
  OEreal min_hail_Zdr_diff;
  OEreal Zdr_prehail_error_factor;
  OEreal prehail_var_ln_a_apriori;
  OEreal distribution_mu;
  OEreal distribution_min_D0;
  OEreal distribution_interval_D0;
  int distribution_num_D0;
  char **argv;
  char *input;
  char *lut_filename;
  char *output_suffix;
  char *output;
  char *first_pass_output_suffix;
  char *first_pass_output;
  char *confstring;
  int min_consecutive_hail_gates;
  int n_basis;
  int do_hail_retrieval;
  int save_first_pass;
  int explicit_range_smoothing;
  int use_Zdr;
  int use_phi_dp;
  int argc;
} Settings;

typedef struct {
  OEmatrix Zdr_fwd;
  OEmatrix phi_dp_fwd;
  OEmatrix R;
  OEmatrix lnR_error;
  OEmatrix Ah;
  OEmatrix Av;
  OEmatrix a;
  OEmatrix lna_error;
  OEmatrix hail_fraction;
  OEmatrix hail_fraction_error;
  OEintmatrix hail_location;
  OEvector atten_scaling;
  OEvector atten_scaling_error;
} Retrieval;

#define DEFAULT_SETTINGS {\
   200.0,      /* a_default */ \
   1.5,        /* b */ \
   10.0,       /* radial_decorr_dist */ \
   10.0,       /* azim_decorr_dist */ \
   1.0,        /* var_ln_a_apriori */ \
   3.0,        /* max_ln_a_excursion */ \
   1.0,        /* Zdr_error_factor */ \
   1.0,        /* phi_dp_error_factor */ \
   1.0,        /* atten_scaling */ \
   10.0,       /* max_atten */ \
   -0.1,       /* var_atten_scaling */ \
   0.1,        /* atten_scaling_decorr */ \
   40.0,       /* min_hail_Zh */ \
   2.0,        /* min_hail_Zdr_diff*/ \
   10.0,       /* Zdr_prehail_error_factor */ \
   1.0,        /* prehail_var_ln_a_apriori */ \
   5,          /* distribution_mu */ \
   0.1,        /* distribution_min_D0 */ \
   0.1,        /* distribution_interval_D0 */ \
   60,         /* distribution_num_D0 */ \
   NULL,       /* Command-line arguments: argv */ \
   NULL,       /* input (filename) */ \
   NULL,       /* lut_filename */ \
   "out",      /* output_suffix */ \
   NULL,       /* output (filename) */ \
   "firstpass",/* first_pass_output_suffix */ \
   NULL,       /* first_pass_output (filename) */ \
   NULL,       /* confstring */ \
   4,          /* min_consecutive_hail_gates*/ \
   10,         /* n_basis */ \
   0,          /* do_hail_retrieval */ \
   0,          /* save_first_pass */ \
   0,          /* explicit_range_smoothing */ \
   0,          /* use Zdr */ \
   0,          /* use phi_dp */ \
   0           /* Number of command-line arguments: argc */ \
}

int load_scattering(char *filename, ScatteringData *scattering);
void free_scattering(ScatteringData *scattering);

int create_scattering_lut(ScatteringData *scattering,
			  int nD0, OEreal D0start, OEreal dD0,
			  OEreal mu, ScatteringLUT *lut);
void free_scattering_lut(ScatteringLUT *lut);

int forward_model(ScatteringLUT *lut, int n, OEreal drange, OEreal *range,
		  OEreal *temperature,
		  OEreal *Zh_lin, OEreal *ln_a, OEreal b,
		  OEreal *hail_fraction,
		  OEreal atten_scaling, OEreal max_atten,
		  /* Output */
		  OEreal *Zdr, OEreal *phi_dp,
		  OEreal *Ah, OEreal *Av,
		  OEreal **dZdr_dlna, OEreal **dphi_dp_dlna,
		  OEreal **dAh_dlna, OEreal **dAv_dlna,
		  OEreal **dZdr_dhf, OEreal **dphi_dp_dhf,
		  OEreal **dAh_dhf, OEreal **dAv_dhf,
		  OEreal *dZdr_das, OEreal *dphi_dp_das,
		  OEreal *dAh_das, OEreal *dAv_das);

#define clear_struct(obj) memset(&obj, 0, sizeof(obj))
int load_scan(char *nc_filename, Scan *scan);
int save_retrieval(char *filename, Retrieval *retrieval, 
		   Scan *scan, Settings *settings);

int zdr_phidp(Settings *settings, ScatteringLUT *lut, Scan *scan,
	      Retrieval *retrieval);

int find_hail(Settings *settings, Scan *scan, Retrieval *retrieval);
int load_settings(int argc, char **argv, Settings *settings);
