#include <stdio.h>
#include <math.h>
#include "zdr_phidp.h"

#define A_PRIORI 0
#define KALMAN_FOREWARD 1
#define KALMAN_BACKWARD 2

static
int
add_spatial_decorrelation(OEmatrix *target, OEreal azim_decorr_factor,
			  OEvector *range, OEreal azim1, OEreal azim2)
{
  OEreal drange, range0, dazim = fabs(azim1 - azim2);
  int i, n = target->n;
  if (n != target->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Spatial decorrelation covariance matrix must be square");
  }
  if (dazim > 360.0) {
    dazim = 360.0 - dazim;
  }

  drange = (range->data[range->n-1]-range->data[0]) / n;
  range0 = range->data[0] + 0.5*drange;

  for (i = 0; i < n; i++) {
    target->data[i][i] += azim_decorr_factor * dazim * (range0+drange*i);
  }

  return OE_SUCCESS;
}
			  
int
zdr_phidp(Settings *settings, ScatteringLUT *lut, Scan *scan,
	  Retrieval *retrieval)
{
  OEvector range = OE_NEW_VECTOR;
  OEvector Zh_lin = OE_NEW_VECTOR;
  OEvector temperature = OE_NEW_VECTOR;

  OEvector y = OE_NEW_VECTOR;
  OEvector inv_sy = OE_NEW_VECTOR;

  OEvector y_fwd = OE_NEW_VECTOR;
  OEvector Zdr_fwd = OE_NEW_VECTOR;
  OEvector phi_dp_fwd = OE_NEW_VECTOR;

  OEvector x = OE_NEW_VECTOR;
  OEvector ln_a = OE_NEW_VECTOR;
  OEvector ln_a_full = OE_NEW_VECTOR;

  OEvector xa = OE_NEW_VECTOR;
  OEvector sa = OE_NEW_VECTOR;

  OEmatrix W_full = OE_NEW_MATRIX;
  OEmatrix W = OE_NEW_MATRIX;

  OEvector Ah = OE_NEW_VECTOR;
  OEvector Av = OE_NEW_VECTOR;

  OEmatrix Jacobian = OE_NEW_MATRIX;

  OEmatrix Jacobian_highres = OE_NEW_MATRIX;
  OEmatrix dZdr_dlna = OE_NEW_MATRIX;
  OEmatrix dphi_dp_dlna = OE_NEW_MATRIX;

  OEmatrix Hessian = OE_NEW_MATRIX;
  OEmatrix invHessian = OE_NEW_MATRIX;
  OEvector rhs = OE_NEW_VECTOR;
  OEvector sp = OE_NEW_VECTOR;

  OEmatrix dAh_dlna = OE_NEW_MATRIX;
  OEmatrix dAv_dlna = OE_NEW_MATRIX;

  OEmatrix x_save = OE_NEW_MATRIX;
  OEmatrix Sx_save = OE_NEW_MATRIX;

  OEintvector gate_list = OE_NEW_VECTOR;

  OEreal ln_a_default = log(settings->a_default);

  OEreal last_azim = 0;

  OEreal sec_fm = 0.0, sec_prep = 0.0, sec_solve = 0.0;
  OEreal sec_spline1 = 0.0, sec_spline2 = 0.0;
  OEreal sec_reset = 0.0, sec_save = 0.0;

  int iray = 0;
  int ngates = scan->range.n;
  int nrays = scan->Zh.m;
  int ix;
  int nx = settings->n_basis;
  int ipass = 0;
  int npass = (settings->azim_decorr_factor > 0.0) + 1;

  OEreal drange = scan->range.data[1]-scan->range.data[0];

  OEreal chi_squared_Zdr = 0.0;
  OEreal chi_squared_phi_dp = 0.0;
  int n_total = 0;

  /*  fprintf(stderr, "%d %d\n", ngates, nx);*/
  OE_CHECK(oe_cubic_b_spline(&W_full, ngates, nx));

  /* Reset the output matrices */
  OE_CHECK(oe_reset_matrix(&retrieval->Zdr_fwd, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->phi_dp_fwd, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->R, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->lnR_error, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->Ah, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->Av, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->a, nrays, ngates));

  /* Set the a priori and its error variance */
  OE_CHECK(oe_reset_vector(&xa, nx));
  OE_CHECK(oe_reset_vector(&sa, nx));
  OE_CHECK(oe_reset_vector(&x, nx));
  for (ix = 0; ix < nx; ix++) {
    xa.data[ix] = ln_a_default;
    sa.data[ix] = settings->var_ln_a_apriori;
    //    x.data[ix] = ln_a_default;
  }

  /* If using a Kalman filter, initialise the matrices to save the
     solution and its error covariance at each ray */
  if (npass > 1) {
    OE_CHECK(oe_reset_matrix(&x_save, nrays, nx));
    OE_CHECK(oe_reset_matrix(&Sx_save, nrays*nx, nx));
  }

  OE_CHECK(oe_reset_int_vector(&gate_list, ngates));

  /* Loop through each pass and each ray */
  if (npass > 1) {
    fprintf(stderr, "Forward pass: %d rays", nrays);
  }

  while (ipass < npass) {
    OEreal last_Ah = 0.0, last_Av = 0.0, last_phi_dp = 0.0;
    int igate = 0;
    int iraingate, nraingates = 0;
    int ny;
    int iter, iter_max = 4;

    oe_time_split();
    /* Find the gates containing rain, i.e. finite values in Zh, Zdr
       and phi_dp */
    while (igate < ngates) {
      if (scan->Zh.data[iray][igate] == scan->Zh.missing_value
	  || (scan->is_Zdr 
	      && scan->Zdr.data[iray][igate]
	      == scan->Zdr.missing_value)
	  || (scan->is_phi_dp
	      && scan->phi_dp.data[iray][igate]
	      == scan->phi_dp.missing_value)) {
	igate++;
      }
      else {
	gate_list.data[nraingates] = igate;
	nraingates++;
	igate++;
      }
    }
    gate_list.n = nraingates;

    //    fprintf(stderr, "Ray %d of %d: %d rainy gates\n",
    //	    iray+1, nrays, nraingates);
    fprintf(stderr, ".");

    if (nraingates < 0) {
      continue;
    }
    /* Reset the vectors to the appropriate size */
    OE_CHECK(oe_reset_vector(&range, nraingates));
    OE_CHECK(oe_reset_vector(&Zh_lin, nraingates));
    OE_CHECK(oe_reset_vector(&temperature, nraingates));
    OE_CHECK(oe_reset_vector(&ln_a, nraingates));
    ny = nraingates*(scan->is_Zdr+scan->is_phi_dp);
    OE_CHECK(oe_reset_vector(&y, ny));
    OE_CHECK(oe_reset_vector(&y_fwd, ny));
    OE_CHECK(oe_reset_vector(&inv_sy, ny));
    if (scan->is_Zdr) {
      OE_CHECK(oe_link_vector(&Zdr_fwd, nraingates, y_fwd.data));
    }
    if (scan->is_phi_dp) {
      OE_CHECK(oe_link_vector(&phi_dp_fwd, nraingates,
			      y_fwd.data + scan->is_Zdr * nraingates));
    }

    OE_CHECK(oe_reset_vector(&Ah, nraingates));
    OE_CHECK(oe_reset_vector(&Av, nraingates));
    OE_CHECK(oe_reset_vector(&rhs, nx));

    /* Reset the matrices to the appropriate size */
    OE_CHECK(oe_reset_matrix(&Jacobian, ny, nx));
    OE_CHECK(oe_reset_matrix(&Jacobian_highres, ny, nraingates));
    OE_CHECK(oe_reset_matrix(&dAh_dlna, nraingates, nraingates));
    OE_CHECK(oe_reset_matrix(&dAv_dlna, nraingates, nraingates));

    /* Subset matrices where appropriate */
    OE_CHECK(oe_subset_matrix(&W, &W_full, &gate_list, 0, nx-1));
    if (scan->is_Zdr) {
      OE_CHECK(oe_link_matrix(&dZdr_dlna, nraingates, nraingates,
			      Jacobian_highres.data));
    }
    if (scan->is_phi_dp) {
      OE_CHECK(oe_link_matrix(&dphi_dp_dlna, nraingates, nraingates,
		      Jacobian_highres.data + nraingates*scan->is_Zdr));
    }

    /* Fill the vectors */
    for (iraingate = 0; iraingate < nraingates; iraingate++) {
      igate = gate_list.data[iraingate];
      range.data[iraingate] = scan->range.data[igate];
      Zh_lin.data[iraingate] = pow(10.0, 0.1*scan->Zh.data[iray][igate]);
      temperature.data[iraingate] = scan->temperature.data[iray][igate];
      if (scan->is_Zdr) {
	y.data[iraingate] = scan->Zdr.data[iray][igate];
	inv_sy.data[iraingate] 
	  = 1.0 / (scan->Zdr_error.data[iray][igate] 
		   * scan->Zdr_error.data[iray][igate]);
      }
      if (scan->is_phi_dp) {
	y.data[iraingate + nraingates*scan->is_Zdr]
	  = scan->phi_dp.data[iray][igate];
	inv_sy.data[iraingate + nraingates*scan->is_Zdr] 
	  = 1.0 / (scan->phi_dp_error.data[iray][igate] 
		   * scan->phi_dp_error.data[iray][igate]);
      }
    }

    /* Set matrices to zero */
    OE_CHECK(oe_zero_matrix(&Jacobian_highres));
    OE_CHECK(oe_zero_matrix(&Jacobian));
    OE_CHECK(oe_zero_matrix(&dAh_dlna));
    OE_CHECK(oe_zero_matrix(&dAv_dlna));
    
    if (nx > 1) {
      OE_CHECK(oe_reset_smoothness_constraint(nx));
      OE_CHECK(oe_register_smoothness_constraint(settings->range_decorr_factor,
						 0, nx-1));
    }

    OE_CHECK(oe_register_diagonal_a_priori_i(A_PRIORI, &xa, &sa));

    //    OE_CHECK(oe_print_a_priori_status(stderr));

    if (npass > 1 && invHessian.n > 0) {
      /* Add Kalman terms */
      OE_CHECK(add_spatial_decorrelation(&invHessian,
					 settings->azim_decorr_factor,
					 &scan->range,
					 scan->azim.data[iray], last_azim));
      //      OE_CHECK(oe_register_a_priori_i(KALMAN_FOREWARD, &x, &invHessian));
      /*
      OE_CHECK(oe_reset_vector(&sp, invHessian.n));
      for (ix = 0; ix < nx; ix++) {
	sp.data[ix] = invHessian.data[ix][ix];
      }
      OE_CHECK(oe_register_diagonal_a_priori_i(KALMAN_FOREWARD, &x, &sp));
      */
      //      OE_CHECK(oe_print_a_priori_status(stderr));
      
      OE_WARN(oe_register_a_priori_i(KALMAN_FOREWARD, &x, &invHessian));
      if (oe_error_code) {
	fprintf(stderr, "Saving erroneous error covariance\n");
	OE_CHECK(oe_save_matrix("Sx_error.dat", &invHessian, "% Sx"));
	exit(oe_error_code);
      }
      
      if (ipass > 0 && iray > 0) {
	OEvector x_p = OE_NEW_VECTOR;
	OEmatrix Sx_p = OE_NEW_MATRIX;

	OE_CHECK(oe_link_vector(&x_p, nx, x_save.data[iray-1]));
	OE_CHECK(oe_link_matrix(&Sx_p, nx, nx,
				Sx_save.data + (iray-1)*nx));
	OE_CHECK(add_spatial_decorrelation(&Sx_p,
					   settings->azim_decorr_factor,
					   &scan->range,
					   scan->azim.data[iray],
					   scan->azim.data[iray-1]));
	OE_CHECK(oe_register_a_priori_i(KALMAN_BACKWARD, &x_p, &Sx_p));
      }
      else {
	OE_CHECK(oe_register_a_priori_i(KALMAN_BACKWARD, NULL, NULL));
      }
    }
    else {
      OE_CHECK(oe_register_a_priori_i(KALMAN_FOREWARD, NULL, NULL));
      OE_CHECK(oe_register_a_priori_i(KALMAN_BACKWARD, NULL, NULL));
    }

    //    OE_CHECK(oe_reset_vector(&x, nx));
    OE_CHECK(oe_reset_matrix(&Hessian, nx, nx));
    OE_CHECK(oe_reset_matrix(&invHessian, nx, nx));
    /* First guess */
    for (ix = 0; ix < nx; ix++) {
      x.data[ix] = ln_a_default;
    }

    sec_reset += oe_time_split();

    //    fprintf(stderr, "Iterating");
    for (iter = 0; iter < iter_max; iter++) {
      //      oe_print_row_vector(stderr, &x, "% x");
      //      fprintf(stderr, ".");

      oe_time_split();
      OE_CHECK(oe_band_diagonal_x_vector(&W, &x, &ln_a));
      sec_spline1 += oe_time_split();      

      OE_CHECK(forward_model(lut, nraingates, drange, range.data,
			     temperature.data, Zh_lin.data, ln_a.data,
			     settings->b,
			     /* Output */
			     Zdr_fwd.data, phi_dp_fwd.data, Ah.data, Av.data,
			     dZdr_dlna.data, dphi_dp_dlna.data,
			     dAh_dlna.data, dAv_dlna.data));

      sec_fm += oe_time_split();

      OE_CHECK(oe_matrix_x_band_diagonal(&Jacobian_highres, &W, &Jacobian));

      sec_spline2 += oe_time_split();

      OE_CHECK(oe_prepare(&x, &y, &y_fwd, &Jacobian, &inv_sy, &Hessian, &rhs));

      sec_prep += oe_time_split();

      if (iter == iter_max-1) {
	break;
      }
      OE_WARN(oe_solve(&Hessian, &rhs));
      if (oe_error_code) {
	sec_solve += oe_time_split();
	fprintf(stderr, "Saving erroneous Hessian and continuing\n");
	OE_CHECK(oe_save_matrix("Hessian_error.dat", &Hessian, "% Hessian"));
	continue;
      }
      sec_solve += oe_time_split();
      //      fprintf(stderr, "\n%g %g %g", x.data[0], x.data[1], x.data[2]);
      for (ix = 0; ix < nx; ix++) {
	x.data[ix] += rhs.data[ix];
      }
    }

    //    OE_CHECK(oe_chi_squared(&y, &y_fwd, &inv_sy, &chi_squared));
    if (ipass+1 == npass) {
      if (scan->is_Zdr) {
	for (iraingate = 0; iraingate < nraingates; iraingate++) {
	  chi_squared_Zdr +=  inv_sy.data[iraingate]
	    * (y.data[iraingate]-y_fwd.data[iraingate])
	    * (y.data[iraingate]-y_fwd.data[iraingate]);
	}
      }
      if (scan->is_phi_dp) {
	for (iraingate = nraingates*scan->is_Zdr;
	     iraingate < nraingates*(1+scan->is_Zdr); iraingate++) {
	  chi_squared_phi_dp += inv_sy.data[iraingate]
	    * (y.data[iraingate]-y_fwd.data[iraingate])
	    * (y.data[iraingate]-y_fwd.data[iraingate]);
	}
      }
      n_total += nraingates;
    }
    //    OE_CHECK(oe_print_row_vector(stderr, &inv_sy, "% inv_sy"));

    //    fprintf(stderr, "%g(%g) ", chi_squared, chi_squared/y.n);

    //    OE_CHECK(oe_save_matrix("Hessian.dat", &Hessian, "% Hessian"));
    //    OE_CHECK(oe_invert_symmetric(&Hessian));
    OE_CHECK(oe_invert(&Hessian, &invHessian));
    /* Save the state if doing Kalman filtering */
    if (npass > 1) {
      OEvector x_p = OE_NEW_VECTOR;
      OEmatrix Sx_p = OE_NEW_MATRIX;
      OE_CHECK(oe_link_vector(&x_p, nx, x_save.data[iray]));
      OE_CHECK(oe_link_matrix(&Sx_p, nx, nx, Sx_save.data + iray*nx));
      OE_CHECK(oe_copy_vector(&x, &x_p));
      OE_CHECK(oe_copy_matrix(&invHessian, &Sx_p));
    }

    /* Put the solution in the retrieval structure */
    igate = 0;
    iraingate = 0;
    while (igate < ngates) {
      if (iraingate < nraingates && gate_list.data[iraingate] == igate) {
	/* A rainy gate */
	if (scan->is_Zdr) {
	  retrieval->Zdr_fwd.data[iray][igate] = Zdr_fwd.data[iraingate];
	}
	if (scan->is_phi_dp) {
	  last_phi_dp = retrieval->phi_dp_fwd.data[iray][igate]
	    = phi_dp_fwd.data[iraingate];
	}
	last_Ah = retrieval->Ah.data[iray][igate] = Ah.data[iraingate];
	last_Av = retrieval->Av.data[iray][igate] = Av.data[iraingate];
	retrieval->R.data[iray][igate] 
	  = pow(Zh_lin.data[iraingate] * pow(10.0, 0.1*Ah.data[iraingate])
		* exp(-ln_a.data[iraingate]), 1/settings->b);
	//	retrieval->lnR_error.data[iray][igate]
	//	  = invHessian.data[iraingate][iraingate]/settings->b;
	iraingate++;
      }
      else {
	/* A non-rainy gate */
	retrieval->Zdr_fwd.data[iray][igate] = last_Av-last_Ah;
	retrieval->R.data[iray][igate] = 0.0;
	//	retrieval->lnR_error.data[iray][igate] = 0.0;
	retrieval->phi_dp_fwd.data[iray][igate] = last_phi_dp;
	retrieval->Ah.data[iray][igate] = last_Ah;
	retrieval->Av.data[iray][igate] = last_Av;
      }
      igate++;
    }
    /* Now do "a" */
    OE_CHECK(oe_link_vector(&ln_a_full, ngates, retrieval->a.data[iray]));
    OE_CHECK(oe_band_diagonal_x_vector(&W_full, &x, &ln_a_full));
    /* Exponentiate */
    for (igate = 0; igate < ngates; igate++) {
      ln_a_full.data[igate] = exp(ln_a_full.data[igate]);
    }
    /*
      OE_CHECK(oe_save_matrix("Jacobian.dat", &Jacobian, "% Jacobian"));
    OE_CHECK(oe_save_matrix("Jacobian_highres.dat", &Jacobian_highres, "% Jacobian_highres"));
    exit(1);
    */
    sec_save += oe_time_split();

    last_azim = scan->azim.data[iray];
    
    if (ipass == 0) {
      /* First pass: move to next ray */
      if (++iray >= nrays) {
	ipass++;
	iray -= 2;
	if (npass > 1) {
	  fprintf(stderr, "\nBackward pass: %d rays", nrays);
	}
      }      
    }
    else {
      /* Second pass: move to previous ray */
      if (--iray < 0) {
	ipass++;
      }
    }
  }
  fprintf(stderr, "\n");

  if (scan->is_Zdr) {
    fprintf(stderr, "Zdr chi-squared = %g (n = %d, rms = %g)\n",
	    chi_squared_Zdr, n_total, sqrt(chi_squared_Zdr/n_total));
  }
  if (scan->is_Zdr) {
    fprintf(stderr, "phi_dp chi-squared = %g (n = %d, rms = %g)\n",
	    chi_squared_phi_dp, n_total, sqrt(chi_squared_phi_dp/n_total));
  }


  fprintf(stderr, "Time spent\n");
  fprintf(stderr, "   resetting matrices:         %g\n", sec_reset);
  fprintf(stderr, "   initial spline conversion:  %g\n", sec_spline1);
  fprintf(stderr, "   forward model:              %g\n", sec_fm);
  fprintf(stderr, "   spline Jacobian conversion: %g\n", sec_spline2);
  fprintf(stderr, "   matrix preparation:         %g\n", sec_prep);
  fprintf(stderr, "   matrix solution:            %g\n", sec_solve);
  fprintf(stderr, "   saving result:              %g\n", sec_save);

  oe_free_vector(&range);
  oe_free_vector(&Zh_lin);
  oe_free_vector(&temperature);

  oe_free_vector(&y);
  oe_free_vector(&inv_sy);

  oe_free_vector(&y_fwd);
  oe_free_vector(&Zdr_fwd);
  oe_free_vector(&phi_dp_fwd);

  oe_free_vector(&x);
  oe_free_vector(&ln_a);
  oe_free_vector(&ln_a_full);

  oe_free_vector(&xa);
  oe_free_vector(&sa);

  oe_free_matrix(&W_full);
  oe_free_matrix(&W);

  oe_free_vector(&Ah);
  oe_free_vector(&Av);

  oe_free_matrix(&Jacobian);

  oe_free_matrix(&Jacobian_highres);
  oe_free_matrix(&dZdr_dlna);
  oe_free_matrix(&dphi_dp_dlna);

  oe_free_matrix(&Hessian);
  oe_free_matrix(&invHessian);
  oe_free_vector(&rhs);
  oe_free_vector(&sp);

  oe_free_matrix(&dAh_dlna);
  oe_free_matrix(&dAv_dlna);

  oe_free_matrix(&x_save);
  oe_free_matrix(&Sx_save);

  oe_free_int_vector(&gate_list);

  return OE_SUCCESS;
}
