#include <stdio.h>
#include <math.h>
#include "zdr_phidp.h"

#define A_PRIORI 0
#define KALMAN_FOREWARD 1
#define KALMAN_BACKWARD 2

static
int
add_spatial_decorrelation(OEmatrix *target, OEreal azim_decorr_factor,
			  OEvector *range, OEreal azim1, OEreal azim2)
{
  OEreal drange, range0, dazim = fabs(azim1 - azim2);
  int i, n = target->n;
  if (n != target->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Spatial decorrelation covariance matrix must be square");
  }
  if (dazim > 360.0) {
    dazim = 360.0 - dazim;
  }

  drange = (range->data[range->n-1]-range->data[0]) / n;
  range0 = range->data[0] + 0.5*drange;

  for (i = 0; i < n; i++) {
    target->data[i][i] += azim_decorr_factor * dazim * (range0+drange*i);
  }

  return OE_SUCCESS;
}
			  
int
zdr_phidp(Settings *settings, ScatteringLUT *lut, Scan *scan,
	  Retrieval *retrieval)
{
  OEvector range = OE_NEW_VECTOR;
  OEvector Zh_lin = OE_NEW_VECTOR;
  OEvector temperature = OE_NEW_VECTOR;

  OEvector y = OE_NEW_VECTOR;
  OEvector inv_sy = OE_NEW_VECTOR;

  OEvector y_fwd = OE_NEW_VECTOR;
  OEvector Zdr_fwd = OE_NEW_VECTOR;
  OEvector phi_dp_fwd = OE_NEW_VECTOR;

  /* The state vector */
  OEvector x = OE_NEW_VECTOR;
  /* The first part of x: basis function ln(a) */
  OEvector ln_a_basis = OE_NEW_VECTOR;
  /* The second part of x: hail fractions in selected gates */
  OEvector hail_fraction = OE_NEW_VECTOR;
  /* The values of ln(a) at rainy gates */
  OEvector ln_a = OE_NEW_VECTOR;
  /* The values of ln(a) at all gates */
  OEvector ln_a_full = OE_NEW_VECTOR;

  OEvector xa = OE_NEW_VECTOR;
  OEvector inv_sa = OE_NEW_VECTOR;

  OEmatrix W_full = OE_NEW_MATRIX;
  OEmatrix W = OE_NEW_MATRIX;

  OEvector Ah = OE_NEW_VECTOR;
  OEvector Av = OE_NEW_VECTOR;

  OEmatrix Jacobian = OE_NEW_MATRIX;
  OEmatrix Jacobian_highres = OE_NEW_MATRIX;
  OEmatrix Jacobian_rainonly = OE_NEW_MATRIX;

  OEmatrix dZdr_dlna = OE_NEW_MATRIX;
  OEmatrix dphi_dp_dlna = OE_NEW_MATRIX;

  OEmatrix Hessian = OE_NEW_MATRIX;
  OEmatrix invHessian = OE_NEW_MATRIX;
  OEvector rhs = OE_NEW_VECTOR;
  OEvector sp = OE_NEW_VECTOR;

  OEmatrix dAh_dlna = OE_NEW_MATRIX;
  OEmatrix dAv_dlna = OE_NEW_MATRIX;
  OEmatrix dAh_dhf = OE_NEW_MATRIX;
  OEmatrix dAv_dhf = OE_NEW_MATRIX;
  OEmatrix dZdr_dhf = OE_NEW_MATRIX;
  OEmatrix dphi_dp_dhf = OE_NEW_MATRIX;

  OEvector dZdr_das = OE_NEW_VECTOR;
  OEvector dphi_dp_das = OE_NEW_VECTOR;
  OEvector dAh_das = OE_NEW_VECTOR;
  OEvector dAv_das = OE_NEW_VECTOR;

  OEmatrix lna_save = OE_NEW_MATRIX;
  OEmatrix Slna_save = OE_NEW_MATRIX;
  OEmatrix invSkalman = OE_NEW_MATRIX;
  OEvector xkalman = OE_NEW_VECTOR;

  OEintvector gate_list = OE_NEW_VECTOR;
  OEintvector hail_gate_list = OE_NEW_VECTOR;

  OEreal ln_a_default = log(settings->a_default);

  OEreal last_azim = 0;

  OEreal sec_fm = 0.0, sec_prep = 0.0, sec_solve = 0.0;
  OEreal sec_spline1 = 0.0, sec_spline2 = 0.0;
  OEreal sec_reset = 0.0, sec_save = 0.0;

  int iray = 0;
  int ngates = scan->range.n;
  int nmaxhailgates = ngates;
  int nrays = scan->Zh.m;
  int ix;
  int nb = settings->n_basis;
  int ipass = 0;
  int npass = (settings->azim_decorr_factor > 0.0) + 1;

  OEreal drange = scan->range.data[1]-scan->range.data[0];

  OEreal chi_squared_Zdr = 0.0;
  OEreal chi_squared_phi_dp = 0.0;
  int n_total = 0;

  if (settings->do_hail_retrieval) {
    fprintf(stderr, "Zdr/Phidp retrieval with hail: %d rays, %d gates\n",
	    nrays, ngates);
  }
  else {
    fprintf(stderr, "Zdr/Phidp retrieval without hail: %d rays, %d gates\n",
	    nrays, ngates);
    nmaxhailgates = 0;
  }
  OE_CHECK(oe_cubic_b_spline(&W_full, ngates, nb));

  /* Reset the output matrices */
  OE_CHECK(oe_reset_matrix(&retrieval->Zdr_fwd, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->phi_dp_fwd, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->R, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->lnR_error, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->Ah, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->Av, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->a, nrays, ngates));
  if (settings->do_hail_retrieval) {
    OE_CHECK(oe_reset_matrix(&retrieval->Zh_hail, nrays, ngates));
    OE_CHECK(oe_zero_matrix(&retrieval->Zh_hail));
  }

  /* Set the a priori and its error variance */
  OE_CHECK(oe_reset_vector(&xa, nb+nmaxhailgates));
  OE_CHECK(oe_reset_vector(&inv_sa, nb+nmaxhailgates));

  /* Elements 0 to nb-1 are the ln(a) values */
  for (ix = 0; ix < nb; ix++) {
    xa.data[ix] = ln_a_default;
    inv_sa.data[ix] = 1.0 / settings->var_ln_a_apriori;
  }
  /* Elements nb to nb+nmaxhailgates-1 are for hail - no a priori
     here */
  for ( ; ix < nb+nmaxhailgates; ix++) {
    xa.data[ix] = 0.0;
    inv_sa.data[ix] = 1.0/(2.0*2.0);
  }

  /* If using a Kalman filter, initialise the matrices to save the
     solution and its error covariance at each ray */
  if (npass > 1) {
    OE_CHECK(oe_reset_matrix(&lna_save, nrays, nb));
    OE_CHECK(oe_reset_matrix(&Slna_save, nrays*nb, nb));
  }

  OE_CHECK(oe_reset_int_vector(&gate_list, ngates));

  /* Loop through each pass and each ray */
  if (npass > 1) {
    fprintf(stderr, "Forward pass ");
  }

  while (ipass < npass) {
    OEreal last_Ah = 0.0, last_Av = 0.0, last_phi_dp = 0.0;
    int igate = 0;
    int iraingate, nraingates = 0;
    int nx, ny;
    int iter, iter_max = 5;
    int ihailgate = 0, nhailgates = 0;

    oe_time_split();

    /* Find the gates containing rain, i.e. finite values in Zh, Zdr
       and phi_dp */
    while (igate < ngates) {
      if (scan->Zh.data[iray][igate] == scan->Zh.missing_value
	  || (scan->is_Zdr 
	      && scan->Zdr.data[iray][igate]
	      == scan->Zdr.missing_value)
	  || (scan->is_phi_dp
	      && scan->phi_dp.data[iray][igate]
	      == scan->phi_dp.missing_value)) {
	igate++;
      }
      else {
	gate_list.data[nraingates] = igate;
	nraingates++;
	igate++;
      }
    }
    gate_list.n = nraingates;

    if (nraingates < 0) {
      continue;
    }

    /* Hail stuff */
    if (settings->do_hail_retrieval) {
      if (!retrieval->hail_location.data) {
	OE_THROW(OE_FAILURE,
	 "Hail location must be set if hail retrieval is to be performed");
      }
      /* Number of hail gates for this ray is stored in the first
	 column of hail_location */
      nhailgates = retrieval->hail_location.data[iray][0];
    }
    if (nhailgates) {
      fprintf(stderr, "H");
    }
    else {
      fprintf(stderr, ".");
    }


    /* Our state vector contains the nb entries for the ln(a) basis
       function values, and nhailgates entries containing the fraction
       of the reflectivity that is due to hail at each gate that
       contains hail */
    nx = nb + nhailgates;
    /* The a priori must be sized to include null elements for the
       hail properties */
    xa.n = nx;
    inv_sa.n = nx;

    /* Reset the vectors to the appropriate size */
    OE_CHECK(oe_reset_vector(&range, nraingates));
    OE_CHECK(oe_reset_vector(&Zh_lin, nraingates));
    OE_CHECK(oe_reset_vector(&temperature, nraingates));
    OE_CHECK(oe_reset_vector(&ln_a, nraingates));
    ny = nraingates*(scan->is_Zdr+scan->is_phi_dp);
    OE_CHECK(oe_reset_vector(&y, ny));
    OE_CHECK(oe_reset_vector(&y_fwd, ny));
    OE_CHECK(oe_reset_vector(&inv_sy, ny));
    if (scan->is_Zdr) {
      OE_CHECK(oe_link_vector(&Zdr_fwd, nraingates, y_fwd.data));
    }
    if (scan->is_phi_dp) {
      OE_CHECK(oe_link_vector(&phi_dp_fwd, nraingates,
			      y_fwd.data + scan->is_Zdr * nraingates));
    }

    OE_CHECK(oe_reset_vector(&Ah, nraingates));
    OE_CHECK(oe_reset_vector(&Av, nraingates));
    OE_CHECK(oe_reset_vector(&rhs, nx));

    /* hail_gate_list contains a list of the indeces to the list of
       rainy gates that also contain hail */
    if (nhailgates) {
      OE_CHECK(oe_reset_int_vector(&hail_gate_list, nhailgates));
    }

    /* Fill the vectors */
    for (iraingate = 0; iraingate < nraingates; iraingate++) {
      igate = gate_list.data[iraingate];
      range.data[iraingate] = scan->range.data[igate];
      Zh_lin.data[iraingate] = pow(10.0, 0.1*scan->Zh.data[iray][igate]);
      temperature.data[iraingate] = scan->temperature.data[iray][igate];
      if (scan->is_Zdr) {
	y.data[iraingate] = scan->Zdr.data[iray][igate];
	inv_sy.data[iraingate] 
	  = 1.0 / (scan->Zdr_error.data[iray][igate] 
		   * scan->Zdr_error.data[iray][igate]);
      }
      if (scan->is_phi_dp) {
	y.data[iraingate + nraingates*scan->is_Zdr]
	  = scan->phi_dp.data[iray][igate];
	inv_sy.data[iraingate + nraingates*scan->is_Zdr] 
	  = 1.0 / (scan->phi_dp_error.data[iray][igate] 
		   * scan->phi_dp_error.data[iray][igate]);
      }
      if (ihailgate < nhailgates
	  && igate == retrieval->hail_location.data[iray][ihailgate+1]) {
	hail_gate_list.data[ihailgate++] = iraingate;
      }
    }
    if (nhailgates) {
      if (ihailgate != nhailgates) {
	OE_THROW(OE_FAILURE, "Hail gates must also be rainy gates");
      }
      /* Note that forward model requires hail info at same gates as
	 rain */
      OE_CHECK(oe_reset_vector(&hail_fraction, nraingates));
      OE_CHECK(oe_zero_vector(&hail_fraction));
    }

    /* Reset the matrices to the appropriate size */
    OE_CHECK(oe_reset_matrix(&Jacobian, ny, nx));
    OE_CHECK(oe_reset_matrix(&Jacobian_highres, ny, nraingates));
    OE_CHECK(oe_reset_matrix(&dAh_dlna, nraingates, nraingates));
    OE_CHECK(oe_reset_matrix(&dAv_dlna, nraingates, nraingates));

    OE_CHECK(oe_reset_vector(&dAh_das, nraingates));
    OE_CHECK(oe_reset_vector(&dAv_das, nraingates));
    if (scan->is_Zdr) {
      OE_CHECK(oe_reset_vector(&dZdr_das, nraingates));
    }
    if (scan->is_phi_dp) {
      OE_CHECK(oe_reset_vector(&dphi_dp_das, nraingates));
    }

    /* Set matrices to zero */    
    OE_CHECK(oe_zero_matrix(&Jacobian_highres));
    OE_CHECK(oe_zero_matrix(&Jacobian));
    OE_CHECK(oe_zero_matrix(&dAh_dlna));
    OE_CHECK(oe_zero_matrix(&dAv_dlna));

    if (nhailgates) {
      OE_CHECK(oe_reset_matrix(&dAh_dhf, nraingates, nraingates));
      OE_CHECK(oe_reset_matrix(&dAv_dhf, nraingates, nraingates));
      OE_CHECK(oe_zero_matrix(&dAh_dhf));
      OE_CHECK(oe_zero_matrix(&dAv_dhf));
      if (scan->is_Zdr) {
	OE_CHECK(oe_reset_matrix(&dZdr_dhf, nraingates, nraingates));
	OE_CHECK(oe_zero_matrix(&dZdr_dhf));
      }
      if (scan->is_phi_dp) {
	OE_CHECK(oe_reset_matrix(&dphi_dp_dhf, nraingates, nraingates));
	OE_CHECK(oe_zero_matrix(&dphi_dp_dhf));
      }
    }

    /* Subset matrices where appropriate */
    OE_CHECK(oe_subset_matrix(&W, &W_full, &gate_list, 0, nb-1));
    if (scan->is_Zdr) {
      OE_CHECK(oe_link_matrix(&dZdr_dlna, nraingates, nraingates,
		      Jacobian_highres.data));
    }
    if (scan->is_phi_dp) {
      OE_CHECK(oe_link_matrix(&dphi_dp_dlna, nraingates, nraingates,
		      Jacobian_highres.data + nraingates*scan->is_Zdr));
    }
    OE_CHECK(oe_link_matrix(&Jacobian_rainonly, ny, nb, Jacobian.data));

    if (nb > 2) {
      OE_CHECK(oe_reset_smoothness_constraint(nx));
      OE_CHECK(oe_register_smoothness_constraint(settings->range_decorr_factor,
						 0, nb-1));
    }

    if (nhailgates) {
      int ih, ih1 = 0, ih2 = 0, ihailgate1 = 0, ihailgate2 = -1;
      for (ih = 0; ih < nhailgates; ih++) {
	int ihailgate = retrieval->hail_location.data[iray][ih+1];
	if (ihailgate2 < 0) {
	  /* First hail gate */
	  ihailgate1 = ihailgate2 = ihailgate;
	  ih1 = ih2 = ih;
	}
	else if (ihailgate > ihailgate2+1 || ih == nhailgates-1) {
	  /* A gap or the final hail gate: smooth the previous
	     contiguous region */
	  if (ih == nhailgates-1) {
	    ih2 = ih;
	  }
	  if (ihailgate2 > ihailgate1+1) {
	    OE_CHECK(oe_register_pinned_smoothness_constraint(200.0, nb+ih1,
							      nb+ih2));
	    //	    fprintf(stderr, "<%d,%d,%d,%d> ", ihailgate1, ihailgate2, ih1, ih2);
	  }
	  ihailgate1 = ihailgate2 = ihailgate;
	  ih1 = ih2 = ih;
	}
	else {
	  ihailgate2 = ihailgate;
	  ih2 = ih;
	}
      }
    }

    OE_CHECK(oe_register_diagonal_a_priori_invs_i(A_PRIORI, &xa, &inv_sa));

    if (npass > 1 && invHessian.n > 0) {
      /* Add Kalman terms */
      OEmatrix invSlna = OE_NEW_MATRIX;
      OE_CHECK(oe_reset_matrix(&invSkalman, nx, nx));
      OE_CHECK(oe_zero_matrix(&invSkalman));
      OE_CHECK(oe_link_matrix(&invSlna, nb, nb, invSkalman.data));
      OE_CHECK(oe_reset_vector(&xkalman, nx));
      OE_CHECK(oe_zero_vector(&xkalman));
      /* Just use the basis-function part if there was hail in
	 previous ray */
      invHessian.m = invHessian.n = nb;
      OE_CHECK(add_spatial_decorrelation(&invHessian,
					 settings->azim_decorr_factor,
					 &scan->range,
					 scan->azim.data[iray], last_azim));
      OE_CHECK(oe_invert(&invHessian, &invSlna));
      OE_CHECK(oe_copy_vector_part(&ln_a_basis, 0, nb, xkalman.data));

      //      OE_CHECK(oe_register_a_priori_i(KALMAN_FOREWARD, &x, &invHessian));
      /*
      OE_CHECK(oe_reset_vector(&sp, invHessian.n));
      for (ix = 0; ix < nx; ix++) {
	sp.data[ix] = invHessian.data[ix][ix];
      }
      OE_CHECK(oe_register_diagonal_a_priori_i(KALMAN_FOREWARD, &x, &sp));
      */
      //      OE_CHECK(oe_print_a_priori_status(stderr));
      
      OE_WARN(oe_register_a_priori_invs_i(KALMAN_FOREWARD,
					  &xkalman, &invSkalman));
      if (oe_error_code) {
	fprintf(stderr, "Saving erroneous inverse error covariance\n");
	OE_CHECK(oe_save_matrix("invSx_error.dat", &invSkalman, "% invSx"));
	exit(oe_error_code);
      }
      
      if (ipass > 0 && iray > 0) {
	OEmatrix Slna = OE_NEW_MATRIX;
	OE_CHECK(oe_link_matrix(&Slna, nb, nb,
				Slna_save.data + (iray-1)*nb));
	OE_CHECK(add_spatial_decorrelation(&Slna,
					   settings->azim_decorr_factor,
					   &scan->range,
					   scan->azim.data[iray],
					   scan->azim.data[iray-1]));
	OE_CHECK(oe_invert(&Slna, &invSlna));
	memcpy(xkalman.data, lna_save.data[iray], nb*sizeof(OEreal));
	OE_CHECK(oe_register_a_priori_invs_i(KALMAN_BACKWARD,
					     &xkalman, &invSkalman));
      }
      else {
	OE_CHECK(oe_register_a_priori_i(KALMAN_BACKWARD, NULL, NULL));
      }
    }
    else {
      OE_CHECK(oe_register_a_priori_i(KALMAN_FOREWARD, NULL, NULL));
      OE_CHECK(oe_register_a_priori_i(KALMAN_BACKWARD, NULL, NULL));
    }

    OE_CHECK(oe_reset_vector(&x, nx));
    OE_CHECK(oe_reset_matrix(&Hessian, nx, nx));
    OE_CHECK(oe_reset_matrix(&invHessian, nx, nx));
    /* First guess */
    /* First the ln(a) values */
    for (ix = 0; ix < nb; ix++) {
      x.data[ix] = ln_a_default;
    }
    /* Then the initial hail fraction values */
    for ( ; ix < nx; ix++) {
      x.data[ix] = 0.0;
    }

    sec_reset += oe_time_split();

    //    fprintf(stderr, "Iterating");
    for (iter = 0; iter < iter_max; iter++) {
      //      oe_print_row_vector(stderr, &x, "% x");
      //      fprintf(stderr, ".");

      oe_time_split();
      OE_CHECK(oe_link_vector(&ln_a_basis, nb, x.data));
      OE_CHECK(oe_band_diagonal_x_vector(&W, &ln_a_basis, &ln_a));
      sec_spline1 += oe_time_split();      

      if (nhailgates) {
	/* Forward model with hail */
	/* First copy hail data from x to hail_fraction */
	for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	  hail_fraction.data[hail_gate_list.data[ihailgate]]
	    = x.data[nb+ihailgate];
	}

	/* Now run forward model */
	OE_CHECK(forward_model(lut, nraingates, drange, range.data,
			       temperature.data, Zh_lin.data, ln_a.data,
			       settings->b, hail_fraction.data,
			       settings->atten_scaling, settings->max_atten,
			       /* Output */
			       Zdr_fwd.data, phi_dp_fwd.data, Ah.data, Av.data,
			       dZdr_dlna.data, dphi_dp_dlna.data,
			       dAh_dlna.data, dAv_dlna.data,
			       dZdr_dhf.data, dphi_dp_dhf.data,
			       dAh_dhf.data, dAv_dhf.data,
			       dZdr_das.data, dphi_dp_das.data,
			       dAh_das.data, dAv_das.data));
      }
      else {
	/* Forward model without hail */
	OE_CHECK(forward_model(lut, nraingates, drange, range.data,
			       temperature.data, Zh_lin.data, ln_a.data,
			       settings->b, NULL,
			       settings->atten_scaling, settings->max_atten,
			       /* Output */
			       Zdr_fwd.data, phi_dp_fwd.data, Ah.data, Av.data,
			       dZdr_dlna.data, dphi_dp_dlna.data,
			       dAh_dlna.data, dAv_dlna.data,
			       NULL, NULL, NULL, NULL,
			       dZdr_das.data, dphi_dp_das.data,
			       dAh_das.data, dAv_das.data));
      }

      sec_fm += oe_time_split();

      /* Note that Jacobian_rainonly is a subset of the Jacobian
	 matrix */
      OE_CHECK(oe_matrix_x_band_diagonal(&Jacobian_highres, &W,
					 &Jacobian_rainonly));
      if (nhailgates) {
	/* Copy relevant parts of derivatives with respect to hail
	   fraction into Jacobian */
	if (scan->is_Zdr) {
	  for (iraingate = 0; iraingate < nraingates; iraingate++) {
	    for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	      Jacobian.data[iraingate][nb+ihailgate]
		= dZdr_dhf.data[iraingate][hail_gate_list.data[ihailgate]];
	    }
	  }
	}
	if (scan->is_phi_dp) {
	  for (iraingate = 0; iraingate < nraingates; iraingate++) {
	    for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	      Jacobian.data[iraingate+nraingates*scan->is_Zdr][nb+ihailgate]
		= dphi_dp_dhf.data[iraingate][hail_gate_list.data[ihailgate]];
	    }
	  }
	}
      }
      /*
      if (nhailgates > 10 && iter == 1) {
	OEmatrix *H;
	OE_CHECK(oe_get_smoothness_constraint(&H));
	OE_CHECK(oe_save_matrix("smoothness_constraint.dat", H, "% H"));
	OE_THROW(OE_FAILURE,"Quitting");
      }
      */
      /*
	OE_CHECK(oe_save_matrix("Jacobian.dat", &Jacobian, "% Jacobian"));
	OE_CHECK(oe_save_matrix("Jacobian_highres.dat", &Jacobian_highres, "% Jacobian"));
	OE_CHECK(oe_save_matrix("dZdr_dhf.dat", &dZdr_dhf, "% dZdr_dhf"));
	OE_CHECK(oe_save_matrix("dphi_dp_dhf.dat", &dphi_dp_dhf, "% dphi_dp_dhf"));
	OE_CHECK(oe_save_matrix("dAh_dhf.dat", &dAh_dhf, "% dAh_dhf"));
	OE_CHECK(oe_save_matrix("dZdr_dlna.dat", &dZdr_dlna, "% dZdr_dlna"));
	OE_CHECK(oe_save_matrix("dphi_dp_dlna.dat", &dphi_dp_dlna, "% dphi_dp_dlna"));
	OE_CHECK(oe_save_matrix("dAh_dlna.dat", &dAh_dlna, "% dAh_dlna"));
	OE_THROW(OE_FAILURE,"Quitting");
      }
      */
      sec_spline2 += oe_time_split();

      OE_CHECK(oe_prepare(&x, &y, &y_fwd, &Jacobian, &inv_sy, &Hessian, &rhs));

      sec_prep += oe_time_split();

      if (iter == iter_max-1) {
	break;
      }
      OE_WARN(oe_solve(&Hessian, &rhs));
      if (oe_error_code) {
	sec_solve += oe_time_split();
	fprintf(stderr, "Error at ray %d iteration %d\n"
		"Saving erroneous Hessian and continuing\n",
		iray, iter);
	OE_CHECK(oe_save_matrix("Hessian_error.dat", &Hessian, "% Hessian"));
	continue;
      }
      sec_solve += oe_time_split();
      //      fprintf(stderr, "\n%g %g %g", x.data[0], x.data[1], x.data[2]);
      for (ix = 0; ix < nx; ix++) {
	x.data[ix] += rhs.data[ix];
      }
      if (nhailgates) {
	/* Check that hail fraction is within bounds */
	for (ix = nb; ix < nx; ix++) {
	  if (x.data[ix] < 0.00001) {
	    x.data[ix] = 0.00001;
	  }
	  else if (x.data[ix] > 0.99) {
	    /* Use 0.99 in case of division by zero somewhere */
	    x.data[ix] = 0.99;
	  }
	}
      }
    }

    //    OE_CHECK(oe_chi_squared(&y, &y_fwd, &inv_sy, &chi_squared));
    if (ipass+1 == npass) {
      /* If this is the backward (second) pass, calculate chi-squared
	 for Zdr and phi_dp */
      if (scan->is_Zdr) {
	for (iraingate = 0; iraingate < nraingates; iraingate++) {
	  chi_squared_Zdr +=  inv_sy.data[iraingate]
	    * (y.data[iraingate]-y_fwd.data[iraingate])
	    * (y.data[iraingate]-y_fwd.data[iraingate]);
	}
      }
      if (scan->is_phi_dp) {
	for (iraingate = nraingates*scan->is_Zdr;
	     iraingate < nraingates*(1+scan->is_Zdr); iraingate++) {
	  chi_squared_phi_dp += inv_sy.data[iraingate]
	    * (y.data[iraingate]-y_fwd.data[iraingate])
	    * (y.data[iraingate]-y_fwd.data[iraingate]);
	}
      }
      n_total += nraingates;
    }
    //    OE_CHECK(oe_print_row_vector(stderr, &inv_sy, "% inv_sy"));

    //    fprintf(stderr, "%g(%g) ", chi_squared, chi_squared/y.n);

    //    OE_CHECK(oe_save_matrix("Hessian.dat", &Hessian, "% Hessian"));
    //    OE_CHECK(oe_invert_symmetric(&Hessian));
    OE_CHECK(oe_invert(&Hessian, &invHessian));
    /* Save the state if doing Kalman filtering */
    if (npass > 1) {
      OE_CHECK(oe_copy_vector_part(&x, 0, nb, lna_save.data[iray]));
      OE_CHECK(oe_copy_matrix_part(&invHessian, 0, 0, nb, nb,
				   &Slna_save, iray*nb, 0));
    }

    if (ipass == npass-1) {
    /* Put the solution in the retrieval structure */
    igate = 0;
    iraingate = 0;
    while (igate < ngates) {
      if (iraingate < nraingates && gate_list.data[iraingate] == igate) {
	/* A rainy gate */
	if (scan->is_Zdr) {
	  retrieval->Zdr_fwd.data[iray][igate] = Zdr_fwd.data[iraingate];
	}
	if (scan->is_phi_dp) {
	  last_phi_dp = retrieval->phi_dp_fwd.data[iray][igate]
	    = phi_dp_fwd.data[iraingate];
	}
	last_Ah = retrieval->Ah.data[iray][igate] = Ah.data[iraingate];
	last_Av = retrieval->Av.data[iray][igate] = Av.data[iraingate];
	retrieval->R.data[iray][igate] 
	  = pow(Zh_lin.data[iraingate] * pow(10.0, 0.1*Ah.data[iraingate])
		* exp(-ln_a.data[iraingate]), 1/settings->b);
	//	retrieval->lnR_error.data[iray][igate]
	//	  = invHessian.data[iraingate][iraingate]/settings->b;
	iraingate++;
      }
      else {
	/* A non-rainy gate */
	retrieval->Zdr_fwd.data[iray][igate] = last_Av-last_Ah;
	retrieval->R.data[iray][igate] = 0.0;
	//	retrieval->lnR_error.data[iray][igate] = 0.0;
	retrieval->phi_dp_fwd.data[iray][igate] = last_phi_dp;
	retrieval->Ah.data[iray][igate] = last_Ah;
	retrieval->Av.data[iray][igate] = last_Av;
      }
      igate++;
    }
    /* Now do "a" */
    OE_CHECK(oe_link_vector(&ln_a_full, ngates, retrieval->a.data[iray]));
    OE_CHECK(oe_band_diagonal_x_vector(&W_full, &ln_a_basis, &ln_a_full));

    /* Exponentiate */
    for (igate = 0; igate < ngates; igate++) {
      ln_a_full.data[igate] = exp(ln_a_full.data[igate]);
    }
    if (nhailgates) {
      for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	int igate = retrieval->hail_location.data[iray][ihailgate+1];
	retrieval->Zh_hail.data[iray][igate]
	  = x.data[nb+ihailgate];// * pow(10.0, 0.1*scan->Zh.data[iray][igate]);
      }
    }
    /*
      OE_CHECK(oe_save_matrix("Jacobian.dat", &Jacobian, "% Jacobian"));
    OE_CHECK(oe_save_matrix("Jacobian_highres.dat", &Jacobian_highres, "% Jacobian_highres"));
    exit(1);
    */
    }
    sec_save += oe_time_split();

    last_azim = scan->azim.data[iray];
    
    if (ipass == 0) {
      /* First pass: move to next ray */
      if (++iray >= nrays) {
	ipass++;
	iray -= 2;
	if (npass > 1) {
	  fprintf(stderr, "\nBackward pass ");
	}
      }      
    }
    else {
      /* Second pass: move to previous ray */
      if (--iray < 0) {
	ipass++;
      }
    }
  }
  fprintf(stderr, "\n");

  if (scan->is_Zdr) {
    fprintf(stderr, "Zdr chi-squared = %g (n = %d, rms = %g)\n",
	    chi_squared_Zdr, n_total, sqrt(chi_squared_Zdr/n_total));
  }
  if (scan->is_Zdr) {
    fprintf(stderr, "phi_dp chi-squared = %g (n = %d, rms = %g)\n",
	    chi_squared_phi_dp, n_total, sqrt(chi_squared_phi_dp/n_total));
  }


  fprintf(stderr, "Time spent\n");
  fprintf(stderr, "   resetting matrices:         %g\n", sec_reset);
  fprintf(stderr, "   initial spline conversion:  %g\n", sec_spline1);
  fprintf(stderr, "   forward model:              %g\n", sec_fm);
  fprintf(stderr, "   spline Jacobian conversion: %g\n", sec_spline2);
  fprintf(stderr, "   matrix preparation:         %g\n", sec_prep);
  fprintf(stderr, "   matrix solution:            %g\n", sec_solve);
  fprintf(stderr, "   saving result:              %g\n", sec_save);

  oe_free_vector(&range);
  oe_free_vector(&Zh_lin);
  oe_free_vector(&temperature);

  oe_free_vector(&y);
  oe_free_vector(&inv_sy);

  oe_free_vector(&y_fwd);
  oe_free_vector(&Zdr_fwd);
  oe_free_vector(&phi_dp_fwd);

  oe_free_vector(&x);
  oe_free_vector(&ln_a);
  oe_free_vector(&ln_a_full);

  oe_free_vector(&xa);
  oe_free_vector(&inv_sa);

  oe_free_matrix(&W_full);
  oe_free_matrix(&W);

  oe_free_vector(&Ah);
  oe_free_vector(&Av);

  oe_free_matrix(&Jacobian);

  oe_free_matrix(&Jacobian_highres);
  oe_free_matrix(&dZdr_dlna);
  oe_free_matrix(&dphi_dp_dlna);

  oe_free_matrix(&Hessian);
  oe_free_matrix(&invHessian);
  oe_free_vector(&rhs);
  oe_free_vector(&sp);

  oe_free_matrix(&dAh_dlna);
  oe_free_matrix(&dAv_dlna);

  oe_free_matrix(&dAh_dhf);
  oe_free_matrix(&dAv_dhf);
  oe_free_matrix(&dZdr_dhf);
  oe_free_matrix(&dphi_dp_dhf);

  oe_free_matrix(&lna_save);
  oe_free_matrix(&Slna_save);

  oe_free_int_vector(&gate_list);

  return OE_SUCCESS;
}
